@@ -82,7 +82,7 @@ typedef struct global
 {
 	char *arg;
 	char *xclude;
-	char *location;
+	char *location;      // is this ever used?
 	FILE *xfp;           // exclusion file (see grep_v())
 	size_t datalen;
 	app_private ap;
@@ -93,7 +93,8 @@ typedef struct global
 	int token_print: 1;  // opt -@
 	int source_print: 1; // opt -s
 	int ex_print: 1;     // opt -X
-	int nu: 5;
+	int nu: 3;
+	int ip_version: 2;   // 0=no set, 1=IPv4, 2=IPv6
 	int verbose: 8;
 } global;
 
@@ -154,6 +155,12 @@ typedef struct query_data
 } query_data;
 
 static inline int do_dup(char *p, char *off, char **dest)
+/*
+* Duplicate a non 0-terminated string p.  off points to the first
+* byte off the string; that is, where a 0 would have been is the string
+* were 0-terminated.  The result is 0-terminated, placed in dest.
+* If dest is non NULL, it is freed before overwriting.
+*/
 {
 	char *dup = NULL;
 
@@ -162,8 +169,10 @@ static inline int do_dup(char *p, char *off, char **dest)
 		size_t len = off - p;
 		while (len > 0 && isspace(((unsigned char*)p)[len - 1]))
 			--len;
-		if ((dup = strndup(p, len)) == NULL)
+		if ((dup = malloc(len + 1)) == NULL)
 			return -1;
+		memcpy(dup, p, len);
+		dup[len] = 0;
 	}
 
 	free(*dest);
@@ -297,7 +306,7 @@ static char *rdap_query(global *glo)
 	if (header) header = curl_slist_append(header,
 		"User-Agent: ipqbdb-reporter/" ABUSERDAP_VERSION);
 	res |= curl_easy_setopt(curl, CURLOPT_HTTPHEADER, header);
-	res |= curl_easy_setopt(curl, CURLOPT_TIMEOUT, 600L);
+	res |= curl_easy_setopt(curl, CURLOPT_TIMEOUT, 15L); // Thu 28 Apr 2022, was 600L
 	res |= curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1L);
 	res |= curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1L);
 	res |= curl_easy_setopt(curl, CURLOPT_HEADERFUNCTION, rdap_header);
@@ -317,6 +326,18 @@ static char *rdap_query(global *glo)
 			res |= curl_easy_setopt(curl, CURLOPT_VERBOSE, 1L);
 	}
 
+	if (glo->ip_version)
+	{
+		long ip_resolve;
+		switch (glo->ip_version)
+		{
+			case 1: ip_resolve = CURL_IPRESOLVE_V4; break;
+			case 2: ip_resolve = CURL_IPRESOLVE_V6; break;
+			default: ip_resolve = CURL_IPRESOLVE_WHATEVER; break;
+		}
+		res |= curl_easy_setopt(curl, CURLOPT_IPRESOLVE, ip_resolve);
+	}
+
 	if (res != CURLE_OK)
 	{
 		report_error(&glo->ap, LOG_CRIT,
@@ -371,7 +392,9 @@ static char *rdap_query(global *glo)
 				if (redirect > glo->max_redir || !q.location)
 					break;
 
-				url = q.location;
+				if (redirect > 0)
+					free(url);
+				url = strdup(q.location);
 				continue;
 			}
 
@@ -398,18 +421,20 @@ static char *rdap_query(global *glo)
 		q.buffer = NULL;
 	}
 
+	if (url != glo->arg)
+	{
+		free(url);
+		url = glo->arg;
+	}
 
 cleanup:
 
 	curl_slist_free_all(header);
 	curl_easy_cleanup(curl);
 	free(q.response);
-
-	// free(q.location);
-	if ((glo->location = q.location) != NULL)
-		glo->arg = q.location;	
-
+	free(q.location);
 	free(q.retry_after);
+
 	if (q.buffer)
 	{
 		if (q.b_used)
@@ -959,6 +984,12 @@ int main(int argc, char *argv[])
 				for (unsigned j = 1; (ch = a[j]) != 0; ++j)
 					switch (ch)
 					{
+						case '4':
+							glo.ip_version |= 1;
+							break;
+						case '6':
+							glo.ip_version |= 2;
+							break;
 						case '@':
 							glo.token_print = 1;
 							break;
