@@ -26,6 +26,7 @@ along with Ipqbdb.  If not, see <http://www.gnu.org/licenses/>.
 
 #include <sys/time.h>
 #include <time.h>
+#include <string.h>
 #include <stdlib.h>
 #include <db.h>
 
@@ -33,27 +34,33 @@ along with Ipqbdb.  If not, see <http://www.gnu.org/licenses/>.
 #define IPBQDB_VERSION 1
 
 /*
-* This is the data part of the main database
-* (the key part is the IPv4 address, uint32_t, in network order)
-* optimized assuming sizeof: time_t=8, int=4, double=8, pid_t=4
+* This is the data part of the main database.
+* For IPv4, the key part is the IPv4 address, uint32_t, in network order.
+* For IPv6, the key part is the IPv6, 16 bytes, in network order, where
+* the host part is filled with 1s; this allows to retrieve the range.
+* Record layout is optimized assuming sizeof: time_t=8, int=4, double=8.
 */
 typedef struct ip_data_t
 {
-	double decay;          // >= IPQBDB_NEVER_DECAY for always blocked
+	double decay;               // >= IPQBDB_NEVER_DECAY for always blocked
 	time_t last_update, last_block, created;
-	int probability;       // 0-RAND_MAX probability of being blocked
-	int reason_id;         // index into the reason file
-	int block_cnt;         // number of packets blocked
-	int decay_add_cnt;     // number of threshold upward passes
+	int probability;            // 0-RAND_MAX probability of being blocked
+	int reason_id;              // index into the reason file
+	int block_cnt;              // number of packets blocked
+	int decay_add_cnt;          // number of threshold upward passes
 // --- added since v.0
-	int caught_cnt;        // number of times caught (from ban or parse)
-	uint32_t start_range;  // reserved --must be 0
-	uint16_t rep_ruling;   // added in v. 1.05
-	unsigned char flag;    // reserved --must be 0
-	unsigned char nu[4];   // reserved --must be 0
+	int caught_cnt;             // number of times caught (from ban or parse)
+	uint32_t not_used;          // reserved --must be 0
+	uint16_t rep_ruling;        // repeated ruling, added in v. 1.05
+	unsigned char flag;          // reserved --must be 0
+	unsigned char plen;         // prefix len, only used for IPv6, added in 2.0
+								// 0 means 128, i.e. a single IP
+	unsigned char official_plen; // prefix len found in RDAP, added in 2.0
+	unsigned char nu[2];   // reserved --must be 0
 	unsigned char chk;     // CHK_SIGNATURE, for loose type checking
 } ip_data_t;
 
+
 /*
 * This is the data part of the "bynumber" (primary) description table
 * (the key part is the reason_id number)
@@ -79,22 +86,40 @@ typedef struct bynumber_descr_t
 */
 typedef struct ip_white_t
 {
-	double decay;	 // 0 = never block this ip, otherwise set initial decay
+	double decay;         // 0 = never block this ip, otherwise set initial decay
 	int probability;      // reserved --must be 0
 	int reason_id;        // reserved --must be 0
-	uint32_t start_range; // reserved --must be 0
-	unsigned char nu[11]; // reserved --must be 0
+	time_t expiration;    // record won't be valid thereafter
+	unsigned char plen;   // prefix len, only used for IPv6, added in 2.0
+	                      // 0 means 128, i.e. a single IP
+	unsigned char nu[6];  // reserved --must be 0
 	unsigned char chk;    // CHK_SIGNATURE, for loose type checking
 } ip_white_t;
 
 // used to be here
 #include "report_error.h"
 
+/*
+* Structure to switch names by adding '6'
+*/
+typedef struct switchable_fname
+{
+	char *home;   // possibly NULL
+	unsigned six; // write '6' at this offset, shift the rest
+	char fname[];
+} switchable_fname;
+
+#include "ip_util.h" // for ip_u
+
+void set_ipqbdb_index_factor(double f);
+double get_ipqbdb_index_factor(void);
+int get_ipqbdb_index_factor_precision(void);
+
 void fputs_database_help(void);
-char *database_fname(char const *name, app_private*);
-char *database_fname_with_home(char const *name, app_private *ap,
-	char const *home, int add_slash_if);
-int open_database(char *name, app_private *ap, DB_ENV**, DB**);
+switchable_fname *database_fname(char const *name, app_private*);
+switchable_fname *database_fname_with_home(char const *name, app_private *ap,
+	char **home, int add_slash_if);
+int open_database(switchable_fname *name, app_private *ap, DB_ENV**, DB**, DB**);
 int close_db(DB *db);
 int close_dbenv(DB_ENV *db_env, int cleanup);
 
@@ -110,17 +135,25 @@ enum
 int set_descr_cnt(DB *db1, int descr_id, int is_new_record,
 	ip_data_t const *deldata, time_t const *tloc);
 int open_descrdb(char *fname, DB_ENV *db_env, DB** db1, DB** db2);
-int check_whitelist(DB *, uint32_t ip, double *decay);
+int check_whitelist(void *, ip_u *ip_addr, double *decay);
+int check_whitelist_e(void *, ip_u *ip_addr, double *decay, unsigned char e[16]);
 
-typedef void (*write_block_cb)(uint32_t ip_addr, int reason_id,
+typedef void (*write_block_cb)(ip_u *ip_addr, void*key, int reason_id,
 	int write_block_force,
 	ip_data_t *ip_data, ip_data_t *old_data, DB *db_d);
 enum
 { write_block_force_decay = 1, write_block_force_probability = 2,
 	write_block_force_reason = 4,
 	write_block_force_add_category = 16 };
-int write_block(DB *db, uint32_t ip_addr, double decay, int reason_id,
+int write_block(DB *db, ip_u *ip_addr, double decay, int reason_id,
 	int* /* I/O */ probability, int force, write_block_cb verbose_fn, DB *db_d);
+	
+#if defined SIMULATION_TEST
+void set_simulation_clock(time_t clock);
+#endif
+
+unsigned make_six(switchable_fname *sf);
+unsigned make_four(switchable_fname *sf);
 
 void fputs_initial_count_help(void);
 
