@@ -26,13 +26,14 @@ along with Ipqbdb.  If not, see <http://www.gnu.org/licenses/>.
 
 typedef struct connkill_cmd
 {
-	size_t ip_ndx;  // the replaced argument
-	size_t argc;    // total number of arguments (excluding term. NULL)
-	char *argv[];   // arguments, including command
+	unsigned ip_ndx;  // the replaced argument
+	unsigned argc;    // total number of arguments (excluding term. NULL)
+	char *argv[];     // arguments, including command
 } connkill_cmd;
 
 connkill_cmd* read_connkill_cmd(app_private *ap);
-void run_exec(connkill_cmd *cmd, char *ip, app_private *ap);
+void run_exec(connkill_cmd *cmd, char const *ip, app_private *ap);
+int system_exec(connkill_cmd *cmd, char const *ip, app_private *ap);
 
 #define EXT_EXEC_H_INCLUDED
 #endif
