@@ -39,6 +39,9 @@ static const char debug_status[] =
 #endif
 	"abled";
 
+#define STR1(...) #__VA_ARGS__
+#define STR2(...) STR1(__VA_ARGS__)
+
 int main()
 {
 	int major = 0, minor = 0;
@@ -65,6 +68,7 @@ int main()
 		sizeof(ip_white_t),
 		sizeof(bynumber_descr_t), IPQBDB_DESCR_MAX_LENGTH);
 
+
 	printf(
 		"IPQBDB_RECENT_DROP_TIME %d\n"
 		"IPQBDB_UPDATE_TICK_TIME %d\n"
@@ -72,7 +76,10 @@ int main()
 		"IPQBDB_INITIAL_COUNT %d\n"
 		"IPQBDB_INITIAL_DECAY %g\n"
 		"IPQBDB_PROBABILITY_THRESHOLD %d (RAND_MAX %jd)\n"
-		"IPQBDB_NEVER_DECAY %f\n#\n"
+		"IPQBDB_NEVER_DECAY  %f\n"
+		"IPQBDB_IPV6_RANGES  %s\n"
+		"IPQBDB_INDEX_FACTOR %f\n"
+		"#\n"
 		"IPQBDB_DEBUG %s\n"
 		"max initial-count %d\n",
 		IPQBDB_RECENT_DROP_TIME,
@@ -82,6 +89,8 @@ int main()
 		IPQBDB_INITIAL_DECAY,
 		IPQBDB_PROBABILITY_THRESHOLD, (intmax_t)RAND_MAX,
 		IPQBDB_NEVER_DECAY,
+		STR2(IPQBDB_IPV6_RANGES_COMMA),
+		IPQBDB_INDEX_FACTOR,
 		debug_status,
 		max_initial_count());
 
