@@ -3,7 +3,7 @@
 * written by ale in milano on 15 nov 2008
 * delete and/or list records in a block.db
 
-Copyright (C) 2008-2021 Alessandro Vesely
+Copyright (C) 2008-2023 Alessandro Vesely
 
 This file is part of Ipqbdb.
 
@@ -52,9 +52,10 @@ along with Ipqbdb.  If not, see <http://www.gnu.org/licenses/>.
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
-#include "ipv4_util.h"
+#include "ip_util.h"
 #include "percent_prob.h"
 #include "rehabilitated_prob.h"
+#include "in_range_ip6.h"
 #define INITIAL_COUNT_H_BOTH
 #include "initial_count.h"
 
@@ -164,6 +165,9 @@ typedef struct histog
 } histog;
 
 static int init_histog(histog *h, time_t secs)
+/*
+* secs is min_time (0 by default).
+*/
 {
 	memset(h, 0, sizeof *h);
 	if ((h->his = (int*)malloc(HISLOG_SIZE * sizeof(int))) == NULL)
@@ -366,7 +370,7 @@ static char const err_prefix[] = "ibd-del";
 
 static char *db_block_name = IPQBDB_DATABASE_NAME;
 static char *db_descr_name = IPQBDB_DESCR_DATABASE_NAME;
-static char *ip_address = "0.0.0.0/0";
+static char *ip_address; // no default
 static char *min_time = "0";
 static char *max_time;
 static char *max_probability = "100%";
@@ -377,7 +381,9 @@ static long max_records = 0;
 static int version_opt, help_opt, syslog_opt, cleanup_opt;
 static int only_blocked, non_blocked;
 static int verbose = -1;
-static int list_opt, raw_opt, del_opt, ban_opt, stat_opt, simple_opt;
+static int list_opt, raw_opt, del_opt, ban_opt, stat_opt, simple_opt, range_opt;
+static int four_only, six_only;
+static char* list_opt_sep_string = " ";
 static struct poptOption opttab[] =
 {
 	{"only-blocked", '\0', POPT_ARG_NONE, &only_blocked, 0,
@@ -394,12 +400,18 @@ static struct poptOption opttab[] =
 	"Select records with probability higher than given(*)", "int[%]"},
 	{"max-probability", '\0', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT, &max_probability, 0,
 	"Select records with probability lower than given(*)", "int[%]"},
-	{"ip-addr", 'i', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT, &ip_address, 0,
+	{"ip-addr", 'i', POPT_ARG_STRING, &ip_address, 0,
 	"The single address or range to delete or list, if any", "ip[-last|/cidr]"},
+	{"ipv4", '4', POPT_ARG_NONE, &four_only, 0,
+	"Do IPv4 (redundant if an ip address is given)", NULL},
+	{"ipv6", '6', POPT_ARG_NONE, &six_only, 0,
+	"Do IPv6 (redundant if an ip address is given)", NULL},
 	{"reason", 'r', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT, &reason_string, 0,
 	"Only consider records matching this", "int|string|regex"},
 	{"ls", 'L', POPT_ARG_NONE, &list_opt, 0,
 	"List selected records", NULL},
+	{"ls-field-separator", '\0', POPT_ARG_STRING|POPT_ARGFLAG_SHOW_DEFAULT, &list_opt_sep_string, 0,
+	"Field separator for --ls action", "single char"},
 	{"ls-ban", '\0', POPT_ARG_NONE, &ban_opt, 0,
 	"List according to ibd-ban format", NULL},
 	{"ls-raw", '\0', POPT_ARG_NONE, &raw_opt, 0,
@@ -410,6 +422,8 @@ static struct poptOption opttab[] =
 	"Delete selected records", NULL},
 	{"stats", 'S', POPT_ARG_NONE, &stat_opt, 0,
 	"List by descr stat counters", NULL},
+	{"ranges", '\0', POPT_ARG_NONE, &range_opt, 0,
+	"Report IPv6 ranges, if any", NULL},
 	{"verbose", 'v', POPT_ARG_INT|POPT_ARGFLAG_OPTIONAL, &verbose, 0,
 	"Be verbose", "level"},
 	{"log-syslog", 'l', POPT_ARG_NONE, &syslog_opt, 0,
@@ -540,7 +554,7 @@ static int all_reason_id(DB *db1, DB *db2)
 		data.data = &u;
 		data.ulen = sizeof u;
 		key.flags = data.flags = DB_DBT_USERMEM;
-		while ((rtc = curs->c_get(curs, &key, &data, DB_NEXT)) == 0 &&
+		while ((rtc = curs->get(curs, &key, &data, DB_NEXT)) == 0 &&
 			caught_signal == 0)
 		{
 			int ovec[30]; // traditional 10 back references
@@ -558,13 +572,13 @@ static int all_reason_id(DB *db1, DB *db2)
 			rtc = 0;
 	}
 	if (rtc)
-		db1->err(db1, rtc, curs? "c_get": "cursor");
+		db1->err(db1, rtc, curs? "get": "cursor");
 	if (curs)
 	{
-		int rtc2 = curs->c_close(curs);
+		int rtc2 = curs->close(curs);
 		if (rtc2)
 		{
-			db2->err(db2, rtc2, "c_close");
+			db2->err(db2, rtc2, "close");
 			rtc = -1;
 		}
 	}
@@ -681,9 +695,13 @@ static int get_probability(int *prob, char *probability)
 	return 0;
 }
 
-static int get_ip_address(ip4_range *ip)
+static int get_ip_address(ip_range *ip)
 {
-	int err = parse_ip_address(ip_address, ip, NULL);
+	int err = 0;
+	if (ip_address)
+		err = parse_ip_address(ip_address, ip, NULL);
+	else
+		memset(ip, 0, sizeof *ip);
 	if (err)
 		report_error(&ap, LOG_ERR, "invalid %s: %s (err=%d)\n",
 			parse_ip_invalid_what(err), ip_address, err);
@@ -691,17 +709,30 @@ static int get_ip_address(ip4_range *ip)
 }
 
 static int
-do_del(DB *db, DB*db_d, ip4_range* ip, time_t *secs, int *prob, histog *his)
+do_del(DB *db, DB*db_d, ip_range* ip, time_t *secs, int *prob, histog *his)
 /*
 * If his != NULL, compute the histogram of records exceeding the given min-time
 * and perform no other operation.
+* secs points to two time_t, min_time (default 0) and max_time (default now).
+* prob points to two int, min_probability (default 0) and max_probability (default 100%).
 * Despite its name, this function also does simple listing of the block database.
 */
 {
-	DBC *curs;
-	DBT key, data;
-	uint32_t ip_last = ip->last;
-	uint32_t cur_key = ip->first;
+	assert(db);
+	assert(db_d || his);  // db_d can be NULL if doing histogram
+	assert(ip);
+	assert(secs);
+	assert(prob);
+
+	unsigned const key_size = ip->ip == 4? 4: 16;
+	unsigned char cur_key[key_size];
+	memcpy(cur_key, ip->u.ip_data, key_size);
+	unsigned char ip_last[key_size];
+	memcpy(ip_last, ip->u2.ip_data, key_size);
+
+	// range_opt
+	unsigned int min_plen = 128, max_plen = 0, tot_ranges = 0, tot_plen = 0;
+
 	int rtc, rtc2;
 	unsigned total_selected = 0, total_del = 0, total_delfail = 0, total_recs = 0;
 	time_t const now = time_invoked, last_time = now - secs[0],
@@ -710,12 +741,14 @@ do_del(DB *db, DB*db_d, ip4_range* ip, time_t *secs, int *prob, histog *his)
 	int max_count = ban_opt? max_initial_count(): 0;
 
 	char descr[IPQBDB_DESCR_MAX_LENGTH];
+	int const list_opt_sep = list_opt_sep_string[0];
 
+	DBT key, data;
 	memset(&key, 0, sizeof key);
 	memset(&data, 0, sizeof data);
 
-	key.data = &cur_key;
-	key.ulen = key.size = sizeof cur_key;
+	key.data = cur_key;
+	key.ulen = key.size = key_size;
 	key.flags = data.flags = DB_DBT_USERMEM;
 	rtc = db->get_pagesize(db, &data.ulen);
 	data.ulen *= 32;
@@ -732,7 +765,8 @@ do_del(DB *db, DB*db_d, ip4_range* ip, time_t *secs, int *prob, histog *his)
 	* been modified since we read it, even if it was requested a min
 	* modified time (secs).
 	*/
-	rtc = db->cursor(db, NULL, &curs, 0);
+	DBC *curs;
+	rtc = db->cursor(db, NULL, &curs, DB_CURSOR_BULK);
 	if (rtc)
 	{
 		db->err(db, rtc, "cannot create cursor");
@@ -744,12 +778,17 @@ do_del(DB *db, DB*db_d, ip4_range* ip, time_t *secs, int *prob, histog *his)
 	* outer loop: fetch a bufferful of records, ordered by IP. The SET_RANGE
 	* flag selects the first record that is not less than the key.
 	*/
+	int const have_arg = ip->args != 0;
+	int const print_range_opt = raw_opt || list_opt || ban_opt || simple_opt;
+	int const af = ip->ip == 4? AF_INET: AF_INET6;
+	char print_range[8] = {0};
+	int db_next = have_arg?  DB_SET_RANGE: DB_FIRST;
 	while (caught_signal == 0)
 	{
 		void *p, *retkey, *retdata;
 		size_t retklen, retdlen;
 
-		rtc = curs->c_get(curs, &key, &data, DB_MULTIPLE_KEY|DB_SET_RANGE);
+		rtc = curs->get(curs, &key, &data, DB_MULTIPLE_KEY|db_next);
 		if (rtc)
 		{
 			if (rtc != DB_NOTFOUND)
@@ -764,29 +803,45 @@ do_del(DB *db, DB*db_d, ip4_range* ip, time_t *secs, int *prob, histog *his)
 		{
 			DB_MULTIPLE_KEY_NEXT(p, &data, retkey, retklen, retdata, retdlen);
 
-			/* test for end-of-buffer and end-of-range */
-			if (p == NULL ||
-				(rtc =
-					memcmp(&ip_last, retkey, sizeof ip_last) < 0?
-						DB_NOTFOUND: 0) != 0)
-							break;
+			if (p == NULL)
+			{
+				rtc = DB_NOTFOUND;
+				break;
+			}
 
 			/* test for sanity */
 			ip_data_t *const ip_data = (ip_data_t*)retdata;
-			if (retdlen < sizeof *ip_data || retklen < sizeof cur_key ||
+			if (retdlen < sizeof *ip_data || retklen < key_size ||
 				ip_data->chk != IPQBDB_CHK_SIGNATURE)
 			{
 				report_error(&ap, LOG_ERR, "Unmatched database record: "
 					"data sz=%zu (expected %zu), "
-					"key sz=%zu (expected %zu), chk=%#x (expected %#x)\n",
-					retdlen, sizeof *ip_data, retklen, sizeof cur_key,
+					"key sz=%zu (expected %u), chk=%#x (expected %#x)\n",
+					retdlen, sizeof *ip_data, retklen, key_size,
 					ip_data->chk, IPQBDB_CHK_SIGNATURE);
 				rtc = 1;
 				break;
 			}
 
-			/* mark key as done */
-			cur_key = *(uint32_t*)retkey;
+			/* test for end-of-buffer and end-of-range */
+			if (have_arg)
+			{
+				if (ip_data->plen)
+				{
+					assert(key_size == 16);
+					if (!in_range_ip6(ip_last, retkey, ip_data->plen))
+					{
+						rtc = DB_NOTFOUND;
+						break;
+					}
+				}
+				else if (memcmp(ip_last, retkey, key_size) < 0)
+				{
+					rtc = DB_NOTFOUND;
+					break;
+				}
+			}
+
 			total_recs += 1;
 
 			/* test for specific selection criteria */
@@ -835,6 +890,7 @@ do_del(DB *db, DB*db_d, ip4_range* ip, time_t *secs, int *prob, histog *his)
 
 				if (del_opt)
 				{
+					key.data = retkey;
 					rtc2 = db->del(db, NULL, &key, 0);
 					if (rtc2)
 					{
@@ -848,18 +904,28 @@ do_del(DB *db, DB*db_d, ip4_range* ip, time_t *secs, int *prob, histog *his)
 					}
 				}
 
-				char ntopbuf[INET_ADDRSTRLEN];
+				char ntopbuf[INET6_ADDRSTRLEN];
+				if (print_range_opt)
+				{
+					if (ip_data->plen)
+					{
+						first_in_range((unsigned char *)retkey, ip_data->plen);
+						sprintf(print_range, "/%-3d", ip_data->plen);
+					}
+					else if (list_opt || key_size == 6)
+						strcpy(print_range, "    ");
+					else
+						print_range[0] = 0;
+				}
 
 				if (raw_opt)
 				{
 					if (verbose && !total_selected)
-						printf("%15s %11s %11s %11s %11s %11s %11s %11s %11s %11s\n",
-							"IP", "CREATED", "PROBABILITY", "BLOCKED", "PACKETS",
-							"UPDATED", "DECAY", "THRESHOLD", "CAUGHT", "DESCRIPTION");
+						printf("%11s %11s %11s %11s %11s %11s %11s %11s %11s %15s\n",
+							"CREATED", "PROBABILITY", "BLOCKED", "PACKETS",
+							"UPDATED", "DECAY", "THRESHOLD", "CAUGHT", "DESCRIPTION", "IP");
 
-					printf("%15s %11d %11d %11d %11d %11d %11g %11d %11d %11d\n",
-						inet_ntop(AF_INET, (struct in_addr*)&cur_key,
-							ntopbuf, sizeof ntopbuf),
+					printf("%11d %11d %11d %11d %11d %11g %11d %11d %11d %15s%s\n",
 						(int)ip_data->created,
 						ip_data->probability,
 						(int)ip_data->last_block,
@@ -868,7 +934,9 @@ do_del(DB *db, DB*db_d, ip4_range* ip, time_t *secs, int *prob, histog *his)
 						ip_data->decay,
 						ip_data->decay_add_cnt,
 						ip_data->caught_cnt,
-						ip_data->reason_id);
+						ip_data->reason_id,
+						inet_ntop(af,
+							retkey, ntopbuf, sizeof ntopbuf), print_range);
 				}
 				else if (list_opt || ban_opt)
 				{
@@ -877,24 +945,24 @@ do_del(DB *db, DB*db_d, ip4_range* ip, time_t *secs, int *prob, histog *his)
 						buf3[TIME_STRING_BUFSIZE];
 
 					if (verbose && list_opt && !total_selected)
-						printf("%15s %8s %7s %8s %10s %8s %10s %10s %10s %s\n",
-							"IP", "CREATED", "PROB.", "BLOCKED", "PACKETS",
-							"UPDATED", "DECAY", "THRESHOLD", "CAUGHT", "DESCRIPTION");
+						printf("%8s %7s %8s %10s %8s %10s %10s %10s %15s     %s\n",
+							"CREATED", "PROB.", "BLOCKED", "PACKETS",
+							"UPDATED", "DECAY", "THRESHOLD", "CAUGHT", "IP", "DESCRIPTION");
 
 					get_descr(db_d, ip_data->reason_id, 0, descr, sizeof descr);
 					if (list_opt)
 					{
-						printf("%15s %8s %6.2f%% %8s %10d %8s %10g %10d %10d %s\n",
-							inet_ntop(AF_INET, (struct in_addr*)&cur_key,
-								ntopbuf, sizeof ntopbuf),
-							time_string(ip_data->created, now, buf1),
-							percent_prob(adjusted_prob),
-							time_string(ip_data->last_block, now, buf2),
-							ip_data->block_cnt,
-							time_string(ip_data->last_update, now, buf3),
-							ip_data->decay,
-							ip_data->decay_add_cnt,
-							ip_data->caught_cnt,
+						printf("%8s%c%6.2f%%%c%8s%c%10d%c%8s%c%10g%c%10d%c%10d%c%15s%s%c%s\n",
+							time_string(ip_data->created, now, buf1), list_opt_sep,
+							percent_prob(adjusted_prob), list_opt_sep,
+							time_string(ip_data->last_block, now, buf2), list_opt_sep,
+							ip_data->block_cnt, list_opt_sep,
+							time_string(ip_data->last_update, now, buf3), list_opt_sep,
+							ip_data->decay, list_opt_sep,
+							ip_data->decay_add_cnt, list_opt_sep,
+							ip_data->caught_cnt, list_opt_sep,
+							inet_ntop(af,
+								retkey, ntopbuf, sizeof ntopbuf), print_range, list_opt_sep,
 							descr);
 					}
 
@@ -905,22 +973,33 @@ do_del(DB *db, DB*db_d, ip4_range* ip, time_t *secs, int *prob, histog *his)
 					* if results are sourced more or less at the same time.
 					*/
 					if (ban_opt)
-						printf("ibd-ban --ip-addr=%s --initial-count=%d"
+						printf("ibd-ban --ip-addr=%s%s --initial-count=%d"
 							" --initial-decay=%g --reason='%s'\n",
-							inet_ntop(AF_INET, (struct in_addr*)&cur_key,
-								ntopbuf, sizeof ntopbuf),
+							inet_ntop(af,
+								retkey, ntopbuf, sizeof ntopbuf), print_range,
 							guess_initial_count(adjusted_prob, max_count),
 							ip_data->decay,
 							descr);
 				}
 				else if (simple_opt)
-					printf("%s was caught %d time%s since %s",
-						inet_ntop(AF_INET, (struct in_addr*)&cur_key,
-							ntopbuf, sizeof ntopbuf),
+					printf("%s%s was caught %d time%s since %s",
+						inet_ntop(af,
+							retkey, ntopbuf, sizeof ntopbuf), print_range,
 						ip_data->caught_cnt,
 						ip_data->caught_cnt > 1? "s": "",
 						ctime(&ip_data->created)/* ctime ends with \n */);
 
+				if (range_opt && ip_data->plen > 0)
+				{
+					unsigned plen = ip_data->plen;
+					if (plen < min_plen)
+						min_plen = plen;
+					if (plen > max_plen)
+						max_plen = plen;
+					tot_plen += plen;
+					tot_ranges += 1;
+				}
+
 				total_selected += 1;
 			}
 		}
@@ -928,10 +1007,10 @@ do_del(DB *db, DB*db_d, ip4_range* ip, time_t *secs, int *prob, histog *his)
 		if (rtc)
 			break;
 
-		cur_key = htonl(ntohl(cur_key) + 1);
+		db_next = DB_NEXT;
 	}
 
-	rtc2 = curs->c_close(curs);
+	rtc2 = curs->close(curs);
 	if (rtc2)
 	{
 		db->err(db, rtc2, "cannot close cursor");
@@ -954,10 +1033,15 @@ do_del(DB *db, DB*db_d, ip4_range* ip, time_t *secs, int *prob, histog *his)
 			sprintf(descr, " in %d sec(s)", (int)(end - now));
 		else
 			descr[0] = 0;
-		printf("%u record(s) selected, %u deleted, %u failed deletion(s)%s\n",
-			total_selected, total_del, total_delfail, descr);
+		char const * const what_ip = key_size == 4? "IPv4": "IPv6";
+		printf("%u %s record(s) selected, %u deleted, %u failed deletion(s)%s\n",
+			total_selected, what_ip, total_del, total_delfail, descr);
 	}
 
+	if (range_opt && tot_ranges > 0)
+		printf("%u IPv6 ranges, prefix length %u min, %.2f avg, %u max\n",
+			tot_ranges, min_plen, (double)tot_plen/(double)tot_ranges, max_plen);
+
 	return (rtc != 0 && rtc != DB_NOTFOUND) || total_delfail;
 }
 
@@ -1054,13 +1138,98 @@ static int do_stat(DB *db)
 	return rtc;
 }
 
+static int
+run_adaptive_del(DB *db_block, ip_range *ip, time_t *secs, int *prob)
+/*
+* Set secs[0] in order to match requested record count.
+* Return -1 on error, 0 to cancel del operation, 1 otherwise;
+*/
+{
+	int rtc = -1; // pessimistic
+	histog his;
+	if (init_histog(&his, secs[0]) == 0)
+	{
+		rtc = do_del(db_block, NULL, ip, secs, prob, &his);
+		if (rtc == 0)
+		{
+			if (his.total_recs <= max_records)
+			{
+				if (verbose)
+					printf(
+						"total IPv%c records = %d <= %ld: no deletion needed!\n",
+						ip->ip == 4? '4': '6',
+						his.total_recs, max_records);
+				// rtc = 0: cancel operation
+			}
+			else if (his.total_recs - his.total_selected < max_records)
+			{
+				char const* err =
+					increase_secs(&his, max_records, secs[0]);
+				if (err)
+				{
+					report_error(&ap, LOG_ERR,
+						"increase_secs: %s\n", err);
+					rtc = -1;
+				}
+				else
+				{
+					if (verbose)
+					{
+						printf("total records = %d, selected = %d,",
+							his.total_recs, his.total_selected);
+						if (his.increase)
+						{
+							char before[TIME_STRING_BUFSIZE],
+								after[TIME_STRING_BUFSIZE];
+							time_string(his.now - secs[0], his.now, before);
+							time_string(his.now - secs[0] - his.increase,
+								his.now, after);
+							if (strcmp(before, after))
+								printf(" min age bumped by %jd: %s to %s!\n",
+									(intmax_t)his.increase, before, after);
+							else
+								printf(
+									" min age increased by %jd before %s!\n",
+										(intmax_t)his.increase, before);
+						}
+						else
+							printf(" min age did not vary!\n");
+					}
+					secs[0] += his.increase; // may get >= secs[1]
+					rtc = 1;
+				}
+			}
+			else
+			{
+				if (verbose)
+					printf("total records = %d, selected = %d,"
+						" expected = %d >= %ld: no adjustment possible!\n",
+						his.total_recs, his.total_selected,
+						his.total_recs - his.total_selected,
+							max_records);
+				rtc = 1;
+			}
+		}
+		else
+			rtc = -1; // do-del error
+		free(his.his);
+	}
+
+	if (rtc < 0)
+		report_error(&ap, LOG_CRIT,
+			"adaptive deletion failed failed for IPv%c\n",
+			ip->ip == 4? '4': '6');
+
+	return rtc;
+}
+
 int main(int argc, char const *argv[])
 {
 	static const char optaliases[] = IPQBDB_OPTION_FILE;
 	static char max_max_time[32];
 	int rtc = 0, errs = 0, prob[2] = {0, 0};
 	time_t secs[2] = {0, time(&time_invoked)};
-	ip4_range ip;
+	ip_range ip;
 
 	sprintf(max_max_time, "%ldS", (long)secs[1]);
 	max_time = max_max_time;
@@ -1081,7 +1250,10 @@ int main(int argc, char const *argv[])
 			err_prefix, poptStrerror(rtc));
 		errs = 1;
 	}
-	else
+
+	int any_ls_opt = list_opt || raw_opt || ban_opt || simple_opt;
+
+	if (rtc == -1)
 	{
 		if (poptPeekArg(opt) != NULL)
 		{
@@ -1110,14 +1282,11 @@ int main(int argc, char const *argv[])
 
 		else
 		{
-			if (raw_opt)
-				list_opt = 1;
-
-			if (!list_opt && !del_opt && !ban_opt && !stat_opt && !simple_opt && errs == 0)
+			if (!any_ls_opt && !del_opt && !stat_opt && !range_opt && errs == 0)
 			{
 				fprintf(stdout,
 					"%s: at least one of"
-					" --ls, --ls-raw, --ls-ban, ls-simple, --stats, or --del"
+					" --ls, --ls-raw, --ls-ban, ls-simple, --stats, --ranges or --del"
 					" must be specified\n",
 					err_prefix);
 				errs = 1;
@@ -1133,6 +1302,9 @@ int main(int argc, char const *argv[])
 			}
 		}
 
+		if (four_only && six_only)
+			four_only = six_only = 0;
+
 		if (get_ip_address(&ip) ||
 			get_seconds(&secs[0], min_time) ||
 			get_seconds(&secs[1], max_time) ||
@@ -1140,6 +1312,12 @@ int main(int argc, char const *argv[])
 			get_probability(&prob[0], min_probability) ||
 			get_probability(&prob[1], max_probability))
 				errs = 3;
+		else if ((four_only && ip.ip == 6) || (six_only && ip.ip == 4))
+		{
+			report_error(&ap, LOG_WARNING, "cannot specify address type"
+				" different from address given\n");
+			errs = 3;
+		}
 
 		if (secs[0] >= secs[1] && errs != 3)
 		{
@@ -1166,12 +1344,31 @@ int main(int argc, char const *argv[])
 	poptFreeContext(opt);
 	rtc = 0;
 
+	/*
+	* If no IP is specified, set whether to run for both.
+	*/
+	time_t secs_6 = secs[0];
+	int run_four = 1, run_six = 1;
+	if (range_opt)
+		six_only = 1;
+	if (ip.ip == 0)
+	{
+		if (four_only)
+			ip.ip = 4;
+		else if (six_only)
+			ip.ip = 6;
+	}
+	if (ip.ip == 4)
+		run_six = 0;
+	else if (ip.ip == 6)
+		run_four = 0;
+
 	if (errs)
 		rtc = 1;
 	else
 	{
-		char *block_fname = database_fname(db_block_name, &ap);
-		char *descr_fname = database_fname(db_descr_name, &ap);
+		switchable_fname *block_fname = database_fname(db_block_name, &ap);
+		switchable_fname *descr_fname = database_fname(db_descr_name, &ap);
 
 		if (block_fname == NULL || descr_fname == NULL)
 			rtc = 1;
@@ -1184,109 +1381,93 @@ int main(int argc, char const *argv[])
 		else
 		{
 			DB_ENV *db_env = NULL;
-			DB *db_block = NULL, *db_de1 = NULL, *db_de2 = NULL;
+			DB *db_block = NULL, *db6_block = NULL, *db_de1 = NULL, *db_de2 = NULL;
 
 			setsigs();
 
-			rtc = open_database(block_fname, &ap, &db_env, &db_block);
+			rtc = open_database(block_fname, &ap, &db_env,
+				ip.ip != 6? &db_block: NULL,
+				ip.ip != 4? &db6_block: NULL);
 
 			if (rtc == 0)
-				rtc = open_descrdb(descr_fname, db_env, &db_de1, &db_de2);
+				rtc = open_descrdb(descr_fname->fname, db_env, &db_de1, &db_de2);
 
 			if (rtc == 0 && caught_signal == 0)
 				rtc = all_reason_id(db_de1, db_de2);
 
 			if (rtc == 0 && caught_signal == 0 && max_records > 0)
 			{
-				histog his;
-				if (init_histog(&his, secs[0]) == 0)
+				if (ip.ip == 0)
 				{
-					rtc = do_del(db_block, db_de1, &ip, secs, prob, &his);
-					if (rtc == 0)
-					{
-						if (his.total_recs <= max_records)
-						{
-							if (verbose)
-								printf(
-									"total records = %d <= %ld: no deletion needed!\n",
-									his.total_recs, max_records);
-							del_opt = 0;
-						}
-						else if (his.total_recs - his.total_selected < max_records)
-						{
-							char const* err =
-								increase_secs(&his, max_records, secs[0]);
-							if (err)
-								report_error(&ap, LOG_ERR,
-									"increase_secs: %s\n", err);
-							else
-							{
-								if (verbose)
-								{
-									printf("total records = %d, selected = %d,",
-										his.total_recs, his.total_selected);
-									if (his.increase)
-									{
-										char before[TIME_STRING_BUFSIZE],
-											after[TIME_STRING_BUFSIZE];
-										time_string(his.now - secs[0], his.now, before);
-										time_string(his.now - secs[0] - his.increase,
-											his.now, after);
-										if (strcmp(before, after))
-											printf(" min age bumped by %jd: %s to %s!\n",
-												(intmax_t)his.increase, before, after);
-										else
-											printf(
-												" min age increased by %jd before %s!\n",
-													(intmax_t)his.increase, before);
-									}
-									else
-										printf(" min age did not vary!\n");
-								}
-								secs[0] += his.increase; // may get >= secs[1]
-							}
-						}
-						else
-						{
-							if (verbose)
-								printf("total records = %d, selected = %d,"
-									" expected = %d >= %ld: no adjustment possible!\n",
-									his.total_recs, his.total_selected,
-									his.total_recs - his.total_selected,
-										max_records);
-						}
-					}
-					free(his.his);
+					ip.ip = 4;
+					run_four = run_adaptive_del(db_block, &ip, secs, prob);
+					time_t sav = secs[0];
+					secs[0] = secs_6;
+					ip.ip = 6;
+					run_six = run_adaptive_del(db6_block, &ip, secs, prob);
+					secs_6 = secs[0];
+					secs[0] = sav;
+					ip.ip = 0;
+				}
+				else if (ip.ip == 4)
+				{
+					assert(db_block);
+					run_four = run_adaptive_del(db_block, &ip, secs, prob);
+				}
+				else if (ip.ip == 6)
+				{
+					assert(db6_block);
+					run_six = run_adaptive_del(db6_block, &ip, secs, prob);
+					secs_6 = secs[0];
+				}
+				else
+				{
+					assert(0);
+					run_four = run_six = 0;
 				}
-				else rtc = 0; // ignore errors during adaptive count
 			}
 
-			if (rtc == 0 && caught_signal == 0 &&
-				(list_opt || raw_opt || ban_opt || del_opt || simple_opt))
+			if (rtc == 0 && caught_signal == 0 && (any_ls_opt || del_opt || range_opt))
 			{
-				if (verbose)
+				if (verbose && (any_ls_opt || del_opt))
 				{
-					printf("%s%s%s records in IP range %d.%d.%d.%d-%d.%d.%d.%d,\n"
-						"\t%s%smin age %jd secs, max age %jd secs,"
+					char range[INET_RANGESTRLEN];
+					char secs_six[40];
+					if (ip.args == 0)
+					{
+						if (secs_6 != secs[0])
+							snprintf(secs_six, sizeof secs_six,
+								" (%jd for IPv6)", (intmax_t)secs_6);
+					}
+					else
+						range[0] = 0;
+
+					printf("%s%s%s records in IP range %s\n"
+						"\t%s%smin age %jd secs%s, max age %jd secs,"
 						" min prob %d=%.2f%%, max prob %d=%.2f%%.\n",
 						del_opt? "delete": "",
-						del_opt && (list_opt || raw_opt || ban_opt)? " and ": "",
-						(list_opt || raw_opt || ban_opt || simple_opt)? "list": "",
-						((unsigned char*)&ip.first)[0],
-						((unsigned char*)&ip.first)[1],
-						((unsigned char*)&ip.first)[2],
-						((unsigned char*)&ip.first)[3],
-						((unsigned char*)&ip.last)[0],
-						((unsigned char*)&ip.last)[1],
-						((unsigned char*)&ip.last)[2],
-						((unsigned char*)&ip.last)[3],
+						del_opt && any_ls_opt? " and ": "",
+						any_ls_opt? "list": "",
+						snprint_range(range, sizeof range, &ip),
 						only_blocked? "only blocked, ": "",
 						non_blocked? "non blocked, ": "",
-						(intmax_t)secs[0], (intmax_t)secs[1],
+						(intmax_t)secs[0], secs_six, (intmax_t)secs[1],
 						prob[0], percent_prob(prob[0]),
 						prob[1], percent_prob(prob[1]));
 				}
-				rtc = do_del(db_block, db_de1, &ip, secs, prob, NULL);
+
+				if (run_four > 0)
+				{
+					ip.ip = 4;
+					rtc = do_del(db_block, db_de1, &ip, secs, prob, NULL);
+				}
+
+				if (run_six > 0)
+				{
+					ip.ip = 6;
+					secs[0] = secs_6;
+					rtc |= do_del(db6_block, db_de1, &ip, secs, prob, NULL);
+				}
 			}
 
 			if (rtc == 0 && caught_signal == 0 && stat_opt)
@@ -1299,10 +1480,16 @@ int main(int argc, char const *argv[])
 
 			close_db(db_de2);
 			close_db(db_de1);
-			close_db(db_block);
+			if (db_block)
+				close_db(db_block);
+			if (db6_block)
+				close_db(db6_block);
 
 			close_dbenv(db_env, cleanup_opt);
 		}
+
+		free(block_fname);
+		free(descr_fname);
 	}
 
 	return rtc;
