@@ -65,7 +65,7 @@ along with Ipqbdb.  If not, see <http://www.gnu.org/licenses/>.
 #include "config_names.h"
 #include "dbstruct.h"
 
-#include "ipv4_util2.h"
+#include "ip_util.h"
 #include "initial_count.h"
 #include "percent_prob.h"
 #include "ext_exec.h"
@@ -190,6 +190,7 @@ static int my_setsigs(void)
 
 static int initial_count = IPQBDB_INITIAL_COUNT;
 static double initial_decay = IPQBDB_INITIAL_DECAY;
+static double index_factor = IPQBDB_INDEX_FACTOR;
 static char *socket_name = IPQBDB_PARSE_SOCKET;
 static char *db_block_name = IPQBDB_DATABASE_NAME;
 static char *db_white_name = IPQBDB_WHITE_DATABASE_NAME;
@@ -216,6 +217,8 @@ static struct poptOption opttab[] =
 	"The default time for the block probability to halve (seconds)", "float"},
 	{"initial-count", 'c', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT, &initial_count, 0,
 	"Default initial full probability count", "integer"},
+	{"index-factor", 'f', POPT_ARG_DOUBLE|POPT_ARGFLAG_SHOW_DEFAULT, &index_factor, 0,
+	"Set range index factor", "float"},
 	{"exec-connkill", 'e', POPT_ARG_NONE, &exec_connkill_opt, 0,
 	"Execute connection kill if probability reaches 100%", NULL},
 	{"exit-on-eof", '\0', POPT_ARG_NONE, &exit_on_eof_opt, 0,
@@ -250,7 +253,7 @@ typedef struct re_list_t
 	int reason;
 	int init_prob;
 	int ip_sub; // 0 if named "IP"
-	unsigned int match, errs; // count of matches, errors
+	unsigned int match4, match6, errs; // count of matches, errors
 } re_list_t;
 
 typedef struct initialization_stuff
@@ -264,8 +267,8 @@ typedef struct initialization_stuff
 	int inited;
 
 	char *home_dir;
-	char **db_fnames;
-	size_t fnames_alloc, fnames_count;
+	switchable_fname **db_fnames;
+	unsigned fnames_alloc, fnames_count;
 
 	int line;
 
@@ -283,7 +286,8 @@ typedef struct runtime_stuff
 
 	DB_ENV *db_env;
 	DB *db1, *db2;
-	DB **db, *db_white;
+	DB **db, **db6, *db_white, *db6_white;
+	DBC *db6c_white;
 
 	re_list_t *re_list;
 
@@ -326,16 +330,24 @@ static void clear_run(runtime_stuff *run)
 {
 	if (run->db)
 		for (size_t i = 0; i < run->db_count; ++i)
+		{
 			close_db(run->db[i]);
+			close_db(run->db6[i]);
+		}
 
+	if (run->db6c_white)
+		run->db6c_white->close(run->db6c_white);
 	close_db(run->db_white);
+	close_db(run->db6_white);
 	close_db(run->db1);
 	close_db(run->db2);
 	close_dbenv(run->db_env, !no_cleanup_opt);
 	free(run->db);
+	free(run->db6);
 	free(run->home_dir);
 	free(run->cmd);
 	clear_re_list(run->re_list);
+	free(run->ovector);
 	memset(run, 0, sizeof *run);
 }
 
@@ -353,81 +365,56 @@ static void clear_sock(socket_stuff *sock)
 	sock->fd = -1;
 }
 
-static char*
+static switchable_fname*
 my_database_fname(initialization_stuff *ini, char const *name)
 {
-	char *fname = database_fname_with_home(name, &ap, ini->home_dir, 1);
+	switchable_fname *fname = database_fname_with_home(name, &ap, &ini->home_dir, 1);
 	if (fname)
 	{
-		char *p = strrchr(fname, '/');
-		if (p == NULL || p == fname)
+		if (ini->home_dir[0] == '-')
 		{
-			report_error(&ap, LOG_ERR, "invalid name \"%s\" at line %d in %s\n",
-				fname, ini->line, pcre_name);
+			report_error(&ap, LOG_ERR,
+				"%s has a directory different than %s at line %d in %s\n",
+				fname->fname, ini->home_dir, ini->line, pcre_name);
 			free(fname);
+			ini->home_dir[0] = '/';
 			return NULL;
 		}
-
-		*p = 0;
-		if (ini->home_dir)
-		{
-			// let all db's share an env in their home directory
-			if (strcmp(ini->home_dir, fname) != 0)
-			{
-				*p = '/';
-				report_error(&ap, LOG_ERR,
-					"%s has a directory different than %s at line %d in %s\n",
-					fname, ini->home_dir, ini->line, pcre_name);
-				free(fname);
-				return NULL;
-			}
-		}
-		else if ((ini->home_dir = strdup(fname)) == NULL)
-		{
-			free(fname);
-			return NULL;
-		}
-		*p = '/';
 	}
-	
+
 	return fname;
 }
 
 static void init_inidb(initialization_stuff *ini)
 {
-	char *descr_fname = my_database_fname(ini, db_descr_name);
+	switchable_fname *descr_fname = my_database_fname(ini, db_descr_name);
 
 	if (descr_fname &&
-		open_database(descr_fname, &ap, &ini->db_env, NULL) == 0 &&
-		open_descrdb(descr_fname, ini->db_env, &ini->db1, &ini->db2) == 0)
+		open_database(descr_fname, &ap, &ini->db_env, NULL, NULL) == 0 &&
+		open_descrdb(descr_fname->fname, ini->db_env, &ini->db1, &ini->db2) == 0)
 			ini->inited = 1;
 	caught_signal = 0; // this is tested by get_reason_id()
 	free(descr_fname);
 }
 
-static int add_fname(initialization_stuff *ini, char *fname)
+static int add_fname(initialization_stuff *ini, char *filename)
 {
-	char *dir_name = my_database_fname(ini, fname);
-	if (dir_name)
+	switchable_fname *fname = my_database_fname(ini, filename);
+	if (fname)
 	{
-		size_t i;
-		char *const p = strrchr(dir_name, '/');
-		char *const db_fname = p + 1;
-		assert(p); // IPQBDB_DATABASE_PREFIX must contain a full path
-		if (p == NULL)
-			return -1;
+		unsigned i;
 
-		*p = 0;
 		for (i = 0; i < ini->fnames_count; ++i)
-			if (strcmp(db_fname, ini->db_fnames[i]) == 0)
+			if (strcmp(fname->fname, ini->db_fnames[i]->fname) == 0)
 			{
-				free(dir_name);
+				free(fname);
 				return i;
 			}
 		if (ini->fnames_alloc <= ini->fnames_count)
 		{
-			size_t alloc = 2 * ini->fnames_alloc + 4;
-			char **db_fnames = (char**)calloc(alloc, sizeof(char*));
+			unsigned alloc = 2 * ini->fnames_alloc + 4;
+			switchable_fname **db_fnames =
+				calloc(alloc, sizeof(switchable_fname*));
 			if (db_fnames == NULL)
 				return -1;
 
@@ -435,7 +422,7 @@ static int add_fname(initialization_stuff *ini, char *fname)
 			{
 				if (ini->fnames_count)
 					memcpy(db_fnames, ini->db_fnames,
-						ini->fnames_count * sizeof (char*));
+						ini->fnames_count * sizeof (switchable_fname*));
 				free(ini->db_fnames);
 			}
 			ini->db_fnames = db_fnames;
@@ -444,12 +431,8 @@ static int add_fname(initialization_stuff *ini, char *fname)
 
 		assert(i == ini->fnames_count);
 		ini->fnames_count += 1;
-		if ((ini->db_fnames[i] = strdup(db_fname)) != NULL)
-		{
-			free(dir_name);
-			return i;
-		}
-		free(dir_name);
+		ini->db_fnames[i] = fname;
+		return i;
 	}
 	return -1;
 }
@@ -478,7 +461,7 @@ static int insert_pcre(initialization_stuff *ini,
 	if (ip[0] == '*' && ip[1] == 0)
 	{
 		static const char host_repl[] = "<HOST>";
-		static const char host_expr[] = "[0fF:]*(?P<IP>(:?[0-9]{1,3}\\.){3}[0-9]{1,3})";
+		static const char host_expr[] = "(?P<IP>([A-Fa-f0-9]{0,4}[.:]:?){1,7}[A-Fa-f0-9]{1,4})";
 		char *repl = strstr(re, host_repl);
 		if (repl == NULL || strstr(repl + 1, host_repl) != NULL)
 		{
@@ -809,6 +792,9 @@ static int read_pcre(int daemonize_if_ok, runtime_stuff *run)
 			initial_count);
 	ini.default_ini_prob += 1; // rounded
 
+	if (index_factor != IPQBDB_INDEX_FACTOR)
+		set_ipqbdb_index_factor(index_factor);
+
 	// are we going to daemonize?
 	// daemonize_if_ok can be 0 if called after SIGHUP, so check -D
 	if (no_daemon_opt || no_run_opt)
@@ -897,26 +883,32 @@ static int read_pcre(int daemonize_if_ok, runtime_stuff *run)
 	run->db1 = ini.db1;
 	run->db2 = ini.db2;
 	run->db = (DB**)malloc(ini.fnames_count * sizeof(DB*));
+	run->db6 = (DB**)malloc(ini.fnames_count * sizeof(DB*));
 	run->ovecsize = (ini.max_sub + 1) * 3;
 	run->ovector = (int*)malloc(run->ovecsize * sizeof(int));
 
-	char *white_fname = database_fname(db_white_name, &ap);
+	switchable_fname *white_fname = my_database_fname(&ini, db_white_name);
 
 	if (run->db &&
 		run->ovector &&
 		white_fname &&
-		open_database(white_fname, &ap, &run->db_env, &run->db_white) == 0)
+		open_database(white_fname, &ap, &run->db_env,
+			&run->db_white, &run->db6_white ) == 0)
 	{
 		size_t i;
 		for (i = 0; i < ini.fnames_count; ++i)
-			if (open_database(ini.db_fnames[i], &ap, &run->db_env, &run->db[i]))
-				break;
+			if (open_database(ini.db_fnames[i], &ap, &run->db_env,
+				&run->db[i], &run->db6[i]))
+					break;
 
 		if (i < ini.fnames_count)
 		{
 			close_dbenv(ini.db_env, !no_cleanup_opt);
 			return -1;
 		}
+
+		if (run->db6_white->cursor(run->db6_white, NULL, &run->db6c_white, 0))
+			return -1;
 	}
 
 	free(white_fname);
@@ -937,6 +929,8 @@ static int read_pcre(int daemonize_if_ok, runtime_stuff *run)
 
 	time(&run->start);
 
+	for (unsigned int i = 0; i < ini.fnames_count; ++i)
+		free(ini.db_fnames[i]);
 	free(ini.db_fnames);
 	return errs;
 }
@@ -971,10 +965,10 @@ static int try_read_pcre(runtime_stuff *run)
 	return rtc;
 }
 
-static int get_ip_address(uint32_t* ip, char *dest, size_t dsize,
+static int get_ip_address(ip_u* ip, char *dest, size_t dsize,
 	char const *src, size_t ssize, int ex_line)
 {
-	struct in_addr in;
+	ip_u in;
 
 	assert(src);
 	assert(dest);
@@ -1022,16 +1016,27 @@ static void spawn_exec(char *ip, connkill_cmd *cmd)
 	}
 
 	run_exec(cmd, ip, &ap);
-	_exit(1); // error logged already
 }
 
-static int block_matched(runtime_stuff *run, re_list_t const * const r,
+enum block_matched_err
+{
+	block_matched_unknown,
+	block_matched_ipv4,
+	block_matched_ipv6,
+	block_matched_ok,
+	block_matched_no_address,
+	block_matched_regex_error,
+	block_matched_lookup_white,
+	block_matched_cannot_write,
+	block_matched_no_db
+};
+static enum block_matched_err block_matched(runtime_stuff *run, re_list_t const * const r,
 	char const *str, size_t const len, int const rc)
 {
-	char dest32[32]; // captured IP
-	uint32_t ip_addr;
-	int no_ip_addr = 1;
-	int err = 0;
+	char dest[INET6_ADDRSTRLEN]; // captured IP
+	ip_u ip_addr;
+	int no_ip_addr = 0;
+	enum block_matched_err err = block_matched_unknown;
 
 	if (rc == 0)
 		report_error(&ap, LOG_WARNING,
@@ -1044,12 +1049,15 @@ static int block_matched(runtime_stuff *run, re_list_t const * const r,
 		int rc2 = pcre_get_named_substring(r->code, str, run->ovector,
 				rc > 0? rc: (int)run->ovecsize/3, "IP", &stringptr);
 		if (rc2 > 0 && stringptr != NULL)
-			no_ip_addr = get_ip_address(&ip_addr, dest32, sizeof dest32,
+			no_ip_addr = get_ip_address(&ip_addr, dest, sizeof dest,
 				stringptr, rc2, r->line);
 		else
+		{
 			report_error(&ap, LOG_ERR,
 				"get_substring rc=%d for expr at line %d, subject=%s"/* no\n */,
 				rc2, r->line, str);
+			err = block_matched_regex_error;
+		}
 	}
 	else
 	{
@@ -1060,55 +1068,79 @@ static int block_matched(runtime_stuff *run, re_list_t const * const r,
 			run->ovector[firstndx] < run->ovector[firstndx + 1] &&
 			(unsigned)run->ovector[firstndx + 1] < len)
 		{
-			no_ip_addr = get_ip_address(&ip_addr, dest32, sizeof dest32,
+			no_ip_addr = get_ip_address(&ip_addr, dest, sizeof dest,
 				str + run->ovector[firstndx],
 				run->ovector[firstndx + 1] - run->ovector[firstndx],
 				r->line);
 		}
 		else
+		{
 			report_error(&ap, LOG_ERR,
 				"mismatch index: ip_sub=%d, first=%d, second=%d"
 				" for expr at line %d, subject=%s"/* no\n */,
 				r->ip_sub, run->ovector[firstndx], run->ovector[firstndx + 1],
 				r->line, str);
+			err = block_matched_regex_error;
+		}
 	}
 
 	if (no_ip_addr)
-		err = 1;
+		err = block_matched_no_address;
 	else
 	{
-		double decay;
-		int rtc, dont_block;
-		rtc = dont_block = check_whitelist(run->db_white, ip_addr, &decay);
-		if (rtc == 1)
+		void *db = NULL;
+		if (ip_addr.ip == 4)
 		{
-			decay = r->decay;
-			dont_block = 0;
+			err = block_matched_ipv4;
+			db = run->db_white;
 		}
-		else if (rtc == 0)
+		else if (ip_addr.ip == 6)
 		{
-			if (decay == 0.0)
-				dont_block = 1;
-			if (verbose || dont_block)
-				report_error(&ap, LOG_NOTICE,
-					"IP captured is whitelisted%s: %s\n",
-					dont_block? " and won't be blocked": "", dest32);
+			err = block_matched_ipv6;
+			db = run->db6c_white;
 		}
 		else
-			err = 1;
+			err = block_matched_no_db;
 
-		if (!dont_block)
+		if (db)
 		{
-			int probability = r->init_prob;
-			rtc = write_block(run->db[r->db_ndx], ip_addr, decay,
-				r->reason, &probability,
-				add_category_opt? write_block_force_add_category: 0,
-				verbose? logline_written_block: NULL, run->db1);
+			double decay;
+			int block = 1;
+			int rtc = check_whitelist(db, &ip_addr, &decay);
+			if (rtc == 0) // not found
+			{
+				decay = r->decay;
+			}
+			else if (rtc == 1)
+			{
+				if (decay == 0.0)
+					block = 0;
+				if (verbose || !block)
+					report_error(&ap, LOG_NOTICE,
+						"IP captured is whitelisted%s: %s\n",
+						block? "": " and won't be blocked", dest);
+			}
+			else
+			{
+				err = block_matched_lookup_white;
+				block = 0;
+			}
 
-			if (rtc != 0)
-				err = 1;
-			else if (probability >= RAND_MAX && run->cmd)
-				spawn_exec(dest32, run->cmd);
+			if (block)
+			{
+				int probability = r->init_prob;
+				int const ndx = r->db_ndx;
+				rtc = write_block(
+					ip_addr.ip == 4? run->db[ndx]: run->db6[ndx],
+					&ip_addr, decay, r->reason, &probability,
+					add_category_opt? write_block_force_add_category: 0,
+					verbose? logline_written_block: NULL, run->db1);
+
+				if (rtc != 0)
+					err = block_matched_cannot_write;
+				else if (probability >= RAND_MAX && run->cmd)
+					spawn_exec(dest, run->cmd);
+			}
 		}
 	}
 
@@ -1146,18 +1178,23 @@ static int run_pcre(runtime_stuff *run, char const *str, size_t const len)
 		}
 		else
 		{
-			int err = block_matched(run, r, str, len, rc);
+			enum block_matched_err err = block_matched(run, r, str, len, rc);
 
-			++r->match;
-			if (err)
+			if (err == block_matched_ipv4)
+				++r->match4;
+			else if (err == block_matched_ipv6)
+				++r->match6;
+			else if (err > block_matched_ok)
 			{
 				++r->errs;
 				rtc = -1;
+				break;
 			}
+
 			/*
 			* move matched rule to front, if not already there
 			*/
-			else if (prev)
+			if (prev)
 			{
 				prev->next = r->next;
 				r->next = run->re_list;
@@ -1204,15 +1241,15 @@ static void log_re_list(runtime_stuff *run)
 
 	re_list_t *r = run->re_list;
 	report_error(&ap, LOG_INFO,
-		"%5.5s %5.5s %5.5s %5.5s %5.5s %19.19s %6.6s %5.5s\n",
-		"LINE", "MATCH", "ERRS", "subex",
+		"%5.5s %6.6s %6.6s %5.5s %5.5s %5.5s %19.19s %6.6s %5.5s\n",
+		"LINE", "MATCH4", "MATCH6", "ERRS", "subex",
 		"db", "initial probability", "reason", "decay");
 	while (r)
 	{
 		report_error(&ap, LOG_INFO,
-			"%5d %5u %5u %5d %5d"
+			"%5d %6u %6u %5u %5d %5d"
 			" %11d %6.2f%% %6d %5g\n",
-			r->line, r->match, r->errs, r->ip_sub, r->db_ndx,
+			r->line, r->match4, r->match6, r->errs, r->ip_sub, r->db_ndx,
 			r->init_prob, percent_prob(r->init_prob), r->reason, r->decay);
 		r = r->next;
 	}
