@@ -55,7 +55,13 @@ sort --key 1 --stable |\
 		done
 		if [ ${#report[@]} -gt $minlin ]; then
 			curl -s https://freegeoip.app/csv/$key
-			abuserdap http://rdap.arin.net/bootstrap/ip/1$key
+			readarray -t <<< "$(abuserdap -s http://rdap.arin.net/bootstrap/ip/$key)"
+			if test -z "${MAPFILE[0]}"; then
+				printf '%s not controlled by an abuse team, see %s\n' \
+					"$key" "${MAPFILE[1]}"
+				# if banning for 1 month in end-of-day, don't use -l!
+				# ibd-ban -i $key -c 0 -t 2592000 -r "IP without abuse team"
+			fi
 			printf "    %s\n" "${report[@]}"
 			echo
 			if [ $first_is_old -eq 1 ]; then
