@@ -1,7 +1,7 @@
 /*
 * rehabilitated_prob.h
 
-Copyright (C) 2011-2019 Alessandro Vesely
+Copyright (C) 2011-2021 Alessandro Vesely
 
 This file is part of Ipqbdb.
 
@@ -35,8 +35,13 @@ rehabilitated_prob(unsigned int prob, time_t delta, double decay)
 	assert(delta > IPQBDB_UPDATE_TICK_TIME);
 	assert(decay > 0.0);
 
+	if (prob > RAND_MAX)
+		prob = RAND_MAX;
+
 	double d = (double)prob * exp2(-(double)delta / decay);
-	return isnormal(d)? (int)floor(0.5 + d): 0;
+	if (isnormal(d) && d > 0.0)
+		return (int)floor(0.5 + d);
+	return 0;
 }
 
 
