@@ -22,34 +22,31 @@ along with Ipqbdb.  If not, see <http://www.gnu.org/licenses/>.
 NOTICE:
 =======
 
-FUBAR
-This work has been derived from src/libnetfilter_queue.c and utils/nfqnl_test.c
-of an old libnetfilter_queue distribution. The content of the original files
-has been completely faked, so that the result cannot serve the purpose of the
-original code. This only serves to link a test program that so that it can run
+This only serves to link a test program that so that it can run
 without root privileges and without disturbing the kernel.
 
 HOW THIS WORKS:
 ===============
 
-The main function runs in the client. It is assumed it calls nfq_open as a
-first thing. Sanity checks are run there, it also retrieves an environment
-variable that has the name of an in-file that with test details, and creates
-the nfq_handle structure filled with the data for the test.
+The main function runs in ibd-judge. It is assumed it calls 
+mnl_socket_open as a first thing. Sanity checks are run there.  The 
+function also retrieves an environment variable that has the name of an 
+in-file with test details, and creates the mnl_socket structure filled 
+with the data for the test.
 
-The in-file may contain packets, commands, and comments. Commands are defined
-in cmd_strings.h; the "group" command groups packets. In-file is first read
-on opening nfq_handle, and continues until there are enough packets to be sent.
+The in-file may contain packets, commands, and comments. Commands are 
+defined in cmd_strings.h; the "group" command groups packets. In-file 
+is first read on opening mnl_socket, and continues until there are 
+enough packets to be sent.
 
-The socketpair is used to pass parsed in-file data to the client. A call to
-nfq_handle_packet is expected after sending. That function executes any command
-and then calls all registered callbacks. The noop command is used to yield
-control to the client. Other commands are useful to set test values, e.g. the
-database, in order for the client to take certain actions. The client's
-decision is displayed when it calls nfq_set_verdict or similar functions.
+A call to mnl_socket_recvfrom is expected next. That function executes 
+any command and then calls all registered callbacks. The noop command 
+is used to yield control to ibd-judge. Other commands are useful to set 
+test values, e.g. the database, in order for ibd-judge to take certain 
+actions.
 
-When the in-file terminates, the socketpair is closed, letting the client's
-call to recv terminate with EBADF. nfq_fd will return -123 after that.
+When the in-file terminates, the socketpair is closed, letting 
+ibd-judge's call to recv terminate with EBADF.
 
 */
 
@@ -68,10 +65,22 @@ call to recv terminate with EBADF. nfq_fd will return -123 after that.
 #include <wait.h>
 #include <sys/types.h>
 #include <netinet/in.h>
-#include <sys/socket.h>
-#include <libnetfilter_queue/libnetfilter_queue.h>
+#include <signal.h>
 
-#include "ipv4_util.h"
+
+// this works with .ko
+// from package libnetfilter-queue-dev
+#include <libmnl/libmnl.h>
+#include <linux/netfilter.h>
+#include <linux/netfilter/nfnetlink.h>
+
+#include <linux/types.h>
+#include <linux/netfilter/nfnetlink_queue.h>
+#include <libnetfilter_queue/libnetfilter_queue.h>
+#include <linux/netfilter/nfnetlink_conntrack.h>
+
+
+#include "ip_util.h"
 
 #include <assert.h>
 
@@ -81,22 +90,17 @@ call to recv terminate with EBADF. nfq_fd will return -123 after that.
 #define MY_PREFIX "TESTJUDGE"
 static const char my_prefix[] = MY_PREFIX;
 
-struct nfq_q_handle
-{
-	struct nfq_handle *h;
-	nfq_callback *cb;
-	void *data;
-};
+struct mnl_socket;
 
 struct nfq_data
 {
+	struct mnl_socket *nl;
 	time_t timestamp;
-	union ipv4
+	union ip
 	{
-		struct fields
+		unsigned version_n_what;
+		struct fields_v4
 		{
-			unsigned version:4;
-			unsigned hlen:4;
 			unsigned tos:8;
 			unsigned len:16;
 			unsigned id:16;
@@ -107,27 +111,43 @@ struct nfq_data
 			unsigned chk:16;
 			unsigned source:32;
 			unsigned dest:32;
-		} v;
-		unsigned char payload[20]; // extra header and data would follow
+		} v4;
+		struct fields_v6
+		{
+			unsigned class:4;  // rest of it, not used
+			unsigned label:20;
+			unsigned len:16;
+			unsigned next:8;
+			unsigned ttl:8;
+			unsigned char source[16];
+			unsigned char dest[16];
+		} v6;
+		unsigned char payload[40]; // extra header and data would follow
 	} packet;
 	int packet_id;
-	struct nfqnl_msg_packet_hdr hdr;
+	int current_q;
+	struct nfgenmsg my_nfgenmsg; // for mnl_nlmsg_get_payload()
+
+	struct nfqnl_msg_packet_hdr my_nfqnl_msg_packet_hdr; // NFQA_PACKET_HDR
+
+	struct nfqnl_msg_packet_timestamp my_nfqnl_msg_packet_timestamp; //NFQA_TIMESTAMP
+
 	unsigned char queue[TEST_QUEUE_MAX];
 };
 
-struct nfq_handle
+struct mnl_socket // fubar
 {
-	// q must be the first element, see get_queue_num
-	struct nfq_q_handle q[TEST_QUEUE_MAX];
 	FILE *in, *out;
 	char *cmd_arg[TEST_ARGS_MAX];
 	time_t current;
-	int sv[2]; // socketpair
 	struct nfq_data next_pkt;
+	void *nextbuf;
+	size_t nextbuf_sz;
 	int line;
-	char sv_open, sv_sent;
+	char sv_sent;
 	char have_next_pkt;
 	char verbose;
+	unsigned char q[TEST_QUEUE_MAX];
 	char fname[];
 };
 
@@ -161,19 +181,8 @@ static const char test_cmd_has_arg[] =
 #undef STRING_MACRO
 
 
-int nfq_errno;
 extern int verbose; // in ibd-judge.c
 
-static void close_socketpair(struct nfq_handle *h)
-{
-	if (h->sv_open)
-	{
-		close(h->sv[0]);
-		close(h->sv[1]);
-		h->sv_open = 0;
-	}	
-}
-
 static int parse_command(char **p)
 {
 	char buf[64];
@@ -207,72 +216,24 @@ static int parse_command(char **p)
 	return test_cmd_invalid;
 }
 
-#if 0 // not used
-static time_t parse_time(char **s)
-// convert from yyyy-mm-dd:hh:mm:ss
-// return -1 for mktime error,
-{
-	char *t = NULL;
-	char const *p = *s;
-	unsigned long ul = strtoul(p, &t, 10);
-	if (ul >= 1900 && ul < INT_MAX && *t == '-')
-	{
-		struct tm tm;
-		char *p = t + 1;
-		memset(&tm, 0, sizeof tm);
-		tm.tm_isdst = -1;
-		tm.tm_year = (int)ul - 1900;
-
-		ul = strtoul(p, &t, 10);
-		if (ul < INT_MAX && *t == '-')
-		{
-			tm.tm_mon = (int)ul - 1;
-			p = t + 1;
-			ul = strtoul(p, &t, 10);
-			if (ul < INT_MAX && *t)
-			{
-				tm.tm_mday = (int)ul;
-				p = t + 1;
-				ul = strtoul(p, &t, 10);
-				if (ul < INT_MAX && *t == ':')
-				{
-					tm.tm_hour = (int)ul;
-					p = t + 1;
-					ul = strtoul(p, &t, 10);
-					if (ul < INT_MAX && *t == ':')
-					{
-						tm.tm_min = (int)ul;
-						p = t + 1;
-						ul = strtoul(p, &t, 10);
-						if (ul < INT_MAX)
-						{
-							tm.tm_sec = (int)ul;
-							time_t rtc = mktime(&tm);
-							if (rtc != (time_t)-1)
-							{
-								*s = t;
-								return rtc;
-							}
-						}
-					}
-				}
-			}
-		}
-	}
-	return (time_t)-1;
-}
-#endif // 0
-
-static int parse_addr(char **p, uint32_t *addr)
+static int parse_addr(char **p, ip_u *addr)
 {
 	char *s = *p, *t = NULL;
-	ip4_range range;
+	ip_range range;
 	int rtc = parse_ip_address(s, &range, &t);
-	if (rtc || range.first != range.last ||
-		t == NULL || (*t && !isspace(*(unsigned char*)t)))
+	if (rtc || range.args != 1 || t == NULL ||
+		(*t && !isspace(*(unsigned char*)t)))
 			return -1;
-	
-	*addr = range.first;
+
+	if (range.ipv4_mapped) // is an IPv6 packet
+	{
+		memmove(&range.u.ipv6[12], &range.u.ipv4[0], 4);
+		range.u.ipv6[10] = range.u.ipv6[11] = 0xff;
+		memset(&range.u.ipv6[0], 0, 10);
+		range.ip = 6;
+	}
+
+	memcpy(addr, &range, sizeof *addr);
 	*p = t;
 	return 0;
 }
@@ -285,12 +246,12 @@ static char const* parse_pkt(char *s, struct nfq_data *pkt)
 */
 {
 	assert(isdigit(*(unsigned char*)s));
-	
+
 	int id;
 	unsigned char queue[TEST_QUEUE_MAX];
-	uint32_t from, to;
-	time_t timestamp = 0;	
-	
+	ip_u from, to;
+	time_t timestamp = 0;
+
 	char *t = NULL;
 	unsigned int l = strtoul(s, &t, 0);
 	int i;
@@ -308,7 +269,7 @@ static char const* parse_pkt(char *s, struct nfq_data *pkt)
 			s += 1;
 		
 		l = strtoul(s, &t, 10);
-		if (*t == '.')
+		if (*t == '.' || *t == ':')  // IP address
 			break;
 
 		if (l >= TEST_QUEUE_MAX)
@@ -316,16 +277,19 @@ static char const* parse_pkt(char *s, struct nfq_data *pkt)
 
 		queue[l] = 1;
 	}
-	
+
 	if (parse_addr(&s, &from))
 		return "source IP";
 	
 	while (isspace(*(unsigned char*)s))
 		s += 1;
-		
+
 	if (parse_addr(&s, &to))
 		return "dest IP";
 
+	if (to.ip != from.ip)
+		return "mixed IPv4 and IPv6";
+
 	while (isspace(*(unsigned char*)s))
 		s += 1;
 
@@ -335,20 +299,32 @@ static char const* parse_pkt(char *s, struct nfq_data *pkt)
 		if (*t)
 			return "time";
 	}
-	
+
 	memset(pkt, 0, sizeof *pkt);
 	pkt->timestamp = timestamp;
-	pkt->packet.v.version = 4;
-	pkt->packet.v.hlen = 5;
-	pkt->packet.v.id = (uint16_t)id;
-	pkt->packet.v.ttl = 8;
-	pkt->packet.v.proto = 6;
-	pkt->packet.v.source = from;
-	pkt->packet.v.dest = to;
+	pkt->packet.version_n_what = (from.ip << 4) + 5;
+	if (from.ip == 4)
+	{
+		pkt->packet.v4.id = (uint16_t)id;
+		pkt->packet.v4.ttl = 8;
+		pkt->packet.v4.proto = 6;
+		pkt->packet.v4.source = from.u.ipv4l;
+		pkt->packet.v4.dest = to.u.ipv4l;
+		// pkt->hdr.packet_id = htonl(id);
+		// pkt->hdr.hw_protocol
+		// pkt->hdr.hook
+	}
+	else
+	{
+		pkt->packet.v6.ttl = 8;
+		memcpy(pkt->packet.v6.source, from.u.ipv6, 16);
+		memcpy(pkt->packet.v6.dest, to.u.ipv6, 16);
+		// pkt->packet.v6.class
+		// pkt->packet.v6.label
+		// pkt->packet.v6.len
+		// pkt->packet.v6.next
+	}
 	pkt->packet_id = id;
-	pkt->hdr.packet_id = htonl(id);
-	// pkt->hdr.hw_protocol
-	// pkt->hdr.hook	
 	memcpy(pkt->queue, queue, sizeof pkt->queue);
 	
 	return 0;
@@ -406,7 +382,7 @@ static char *simple_expand(char *arg)
 	return arg;
 }
 
-static int get_free_arg(struct nfq_handle *h)
+static int get_free_arg(struct mnl_socket *h)
 {
 	unsigned i;
 	for (i = 0; i < sizeof h->cmd_arg/sizeof h->cmd_arg[0]; ++i)
@@ -419,13 +395,15 @@ static int get_free_arg(struct nfq_handle *h)
 	return -1;
 }
 
-static int read_test_input(struct nfq_handle *h)
+static int read_test_input(struct mnl_socket *h)
 /*
 * read in-file and send 1 buffer of commands until one pkt is ready,
 * or return -1 for failure.
 */
 {
 	assert(h && h->have_next_pkt == 0);
+	if (h->in == NULL)
+		return -1;
 
 	char buf[1024], *s;
 	char cmdbuf[TEST_CMD_MAX * sizeof(test_cmd)];
@@ -530,7 +508,6 @@ static int read_test_input(struct nfq_handle *h)
 							else if ((h->cmd_arg[ndx] = strdup(s)) == NULL)
 							{
 								fputs("OUT OF MEMORY\n", stderr);
-								close_socketpair(h);
 								return -1;
 							}
 							else
@@ -590,9 +567,14 @@ static int read_test_input(struct nfq_handle *h)
 		}
 	}
 
-	if (!h->have_next_pkt || !h->sv_open)
+	if (feof(h->in))
+	{
+		fclose(h->in);
+		h->in = NULL;
+	}
+
+	if (!h->have_next_pkt)
 	{
-		close_socketpair(h);
 		return -1;
 	}
 
@@ -606,24 +588,29 @@ static int read_test_input(struct nfq_handle *h)
 		memcpy(&cmdbuf[in_cmdbuf], &tc, sizeof tc);
 		in_cmdbuf += sizeof tc;
 	}
-	
-	if (write(h->sv[0], cmdbuf, in_cmdbuf) == -1)
+
+	h->nextbuf = malloc(h->nextbuf_sz = in_cmdbuf);
+	if (h->nextbuf == NULL)
 	{
 		fprintf(stderr,
-			MY_PREFIX ": cannot send through socketpair: %s\n",
-			strerror(errno));
-		close_socketpair(h);
+			MY_PREFIX ": out of memory (%zu)\n", in_cmdbuf);
 		return -1;
 	}
 
+	memcpy(h->nextbuf, cmdbuf, in_cmdbuf);
 	h->sv_sent = 1;
 	return 0;
 }
 
+static void suicide(void)
+{
+	kill(getpid(), SIGTERM);
+}
+
 #if defined __GNUC__
 __attribute__ ((format(printf, 2, 3)))
 #endif
-static int print_out(struct nfq_handle *h, char const *fmt, ...)
+static int print_out(struct mnl_socket *h, char const *fmt, ...)
 {
 	int rtc = 0;
 	if (h->verbose && h->out)
@@ -637,40 +624,45 @@ static int print_out(struct nfq_handle *h, char const *fmt, ...)
 	return rtc;
 }
 
-static void send_noop_packet(struct nfq_handle *h)
+/* use instead of nlmsghdr for configuration */
+struct my_params
 {
-	assert(h && h->sv_open);
+	uint32_t nlmsg_len; // must be first, for cast to nlmsg in nfq_nlmsg_put()
+	uint32_t queue_num;
+	struct mnl_socket* nl;
+	int range;
+	int verdict;
+	int id;
+	uint32_t queue_maxlen;
+	uint32_t mark;
+	uint8_t cmd;
+};
+
+enum my_cmd
+{
+	my_NFQNL_CFG_CMD_BIND,
+	my_NFQNL_CFG_CMD_UNBIND,
+	my_NFQNL_COPY_PACKET,
+	my_NFQNL_MSG_VERDICT,
+	my_NFQNL_MSG_CONFIG_MAXLEN,
+	my_NFQNL_IGNORE_COMMAND
+};
+
 
-	test_cmd tc;
-	memset(&tc, 0, sizeof tc);
-	tc.cmd = test_cmd_noop;
-	if (write(h->sv[0], &tc, sizeof tc) == -1)
-		close_socketpair(h);
-	else
-		h->sv_sent = 1;
-}
 
 /* public interface */
 
-struct nfnl_handle *nfq_nfnlh(struct nfq_handle *h)
+struct mnl_socket *mnl_socket_open(int bus)
+// was struct nfq_handle *nfq_open_nfnl(struct nfnl_handle *nfnlh)
 {
-	(void)h;
-	return NULL; // not used
-}
+	if (bus != NETLINK_NETFILTER)
+	{
+		fprintf(stderr,
+			MY_PREFIX ": unexpected value %d, should be %d\n",
+			bus, NETLINK_NETFILTER);
+		return NULL;
+	}
 
-int nfq_fd(struct nfq_handle *h)
-{
-	return h->sv_open? h->sv[1]: -123;
-}
-
-struct nfq_handle *nfq_open(void)
-{
-	return nfq_open_nfnl(NULL);
-}
-
-struct nfq_handle *nfq_open_nfnl(struct nfnl_handle *nfnlh)
-{
-	struct nfq_handle *h;
 	char *fname = getenv(my_prefix);
 	size_t fname_size = fname? strlen(fname) + 1: 0;
 
@@ -690,7 +682,8 @@ struct nfq_handle *nfq_open_nfnl(struct nfnl_handle *nfnlh)
 			stderr);
 		return NULL;
 	}
-		
+
+	struct mnl_socket *h;
 	h = malloc(sizeof *h + fname_size);
 	if (!h)
 		return NULL;
@@ -704,19 +697,9 @@ struct nfq_handle *nfq_open_nfnl(struct nfnl_handle *nfnlh)
 			fname, strerror(errno));
 		goto error_exit;
 	}
-		
-	if (socketpair(PF_UNIX, SOCK_SEQPACKET, 0, h->sv) == 0)
-	{
-		h->sv_open = 1;
-		if (!read_test_input(h) == 0)
-			goto error_exit;
 
-	} else
-	{
-		fprintf(stderr,
-			MY_PREFIX ": socketpair error: %s\n", strerror(errno));
+	if (!read_test_input(h) == 0)
 		goto error_exit;
-	}
 
 	h->out = stdout;
 	h->verbose = verbose != 1;
@@ -727,15 +710,14 @@ struct nfq_handle *nfq_open_nfnl(struct nfnl_handle *nfnlh)
 	{
 		if (h->in) fclose(h->in);
 		free(h);
+		suicide();
 	}
 
 	return NULL;
-	(void)nfnlh;
 }
 
-int nfq_close(struct nfq_handle *h)
+int mnl_socket_close(struct mnl_socket *h)
 {
-	close_socketpair(h);
 	if (h->out && h->out != stdout) fclose(h->out);
 	if (h->in) fclose(h->in);
 	free(h);
@@ -743,91 +725,287 @@ int nfq_close(struct nfq_handle *h)
 }
 
 /* bind nf_queue from a specific protocol family */
-int nfq_bind_pf(struct nfq_handle *h, u_int16_t pf)
-{
-	(void)h;
-	(void)pf; // should check it is AF_INET
+int mnl_socket_bind(struct mnl_socket *nl, unsigned int groups,
+				  pid_t pid){
+	(void)nl;
+	(void)groups;
+	(void)pid;
 	return 0;
 }
 
-/* unbind nf_queue from a specific protocol family */
-int nfq_unbind_pf(struct nfq_handle *h, u_int16_t pf)
+struct nlmsghdr *nfq_nlmsg_put(char *buf, int type, uint32_t queue_num)
 {
-	(void)h;
-	(void)pf; // should check it is AF_INET
+	struct my_params par;
+	par.nlmsg_len = sizeof par;
+	par.queue_num = queue_num;
+	memcpy(buf, &par, sizeof par);
+
+	return (struct nlmsghdr *)buf;
+	(void)type;
+}
+
+void nfq_nlmsg_cfg_put_cmd(struct nlmsghdr *nlh, uint16_t pf, uint8_t cmd)
+{
+	struct my_params *par = (struct my_params*)nlh;
+	switch (cmd)
+	{
+		case NFQNL_CFG_CMD_BIND:
+			par->cmd = my_NFQNL_CFG_CMD_BIND;
+			break;
+		case NFQNL_CFG_CMD_UNBIND:
+			par->cmd = my_NFQNL_CFG_CMD_UNBIND;
+			break;
+	}
+	(void)pf;
+}
+
+void nfq_nlmsg_cfg_put_params(struct nlmsghdr *nlh, uint8_t mode, int range)
+{
+	struct my_params *par = (struct my_params*)nlh;
+	par->range = range;
+	par->cmd = my_NFQNL_COPY_PACKET;
+	(void)mode;
+}
+
+void nfq_nlmsg_cfg_put_qmaxlen(struct nlmsghdr *nlh, uint32_t queue_maxlen)
+{
+	struct my_params *par = (struct my_params*)nlh;
+	par->cmd = my_NFQNL_MSG_CONFIG_MAXLEN;
+	par->queue_maxlen = queue_maxlen;
+}
+
+void mnl_attr_put_u32(struct nlmsghdr *nlh, uint16_t type,
+				    uint32_t data)
+{
+	struct my_params *par = (struct my_params*)nlh;
+
+	if (type == CTA_MARK)
+		par->mark = data;
+
+	par->cmd = my_NFQNL_IGNORE_COMMAND;
+}
+
+void nfq_nlmsg_verdict_put(struct nlmsghdr *nlh, int id, int verdict)
+{
+	struct my_params *par = (struct my_params*)nlh;
+	par->id = id;
+	par->cmd = my_NFQNL_MSG_VERDICT;
+	par->mark = 0;
+	par->verdict = verdict;
+
+}
+
+struct nlattr *mnl_attr_nest_start(struct nlmsghdr *nlh,
+						 uint16_t type)
+{
+	return NULL;
+	(void)nlh;
+	(void)type;
+}
+
+void mnl_attr_nest_end(struct nlmsghdr *nlh,
+				     struct nlattr *start)
+{
+	(void)nlh;
+	(void)start;
+}
+
+int nfq_nlmsg_parse(const struct nlmsghdr *nlh, struct nlattr **attr)
+{
+	for (int i = 0; i <= NFQA_MAX; ++i)
+		attr[i] = NULL;
+
+	struct nfq_data *pkt = (struct nfq_data*)nlh;
+	attr[NFQA_PAYLOAD] = (struct nlattr*)&pkt->packet;
+	attr[NFQA_PACKET_HDR] = (struct nlattr*)&pkt->my_nfqnl_msg_packet_hdr; // payload
+	memset(&pkt->my_nfqnl_msg_packet_hdr, 0, sizeof pkt->my_nfqnl_msg_packet_hdr);
+	pkt->my_nfqnl_msg_packet_hdr.packet_id = htonl(pkt->packet_id);
+	if (pkt->timestamp)
+	{
+		attr[NFQA_TIMESTAMP] =(struct nlattr*) &pkt->my_nfqnl_msg_packet_timestamp;
+		pkt->my_nfqnl_msg_packet_timestamp.sec = htobe64(pkt->timestamp);
+		pkt->my_nfqnl_msg_packet_timestamp.usec = 0;
+	}
 	return 0;
 }
 
-/* bind this socket to a specific queue number */
-struct nfq_q_handle *nfq_create_queue(struct nfq_handle *h, 
-		u_int16_t num,
-		nfq_callback *cb,
-		void *data)
+void *mnl_nlmsg_get_payload(const struct nlmsghdr *nlh)
 {
+	struct nfq_data *pkt = (struct nfq_data*)nlh;
+	pkt->my_nfgenmsg.res_id = htons(pkt->current_q);
+	return (void *)&pkt->my_nfgenmsg;
+}
+
+uint16_t mnl_attr_get_payload_len(const struct nlattr *attr)
+{
+	return 40;
+	(void)attr;
+}
+
+void *mnl_attr_get_payload(const struct nlattr *attr)
+{
+	return (void*)attr;
+}
+
+int mnl_socket_setsockopt(const struct mnl_socket *nl, int type,
+					void *buf, socklen_t len)
+ // called if netlink_no_enobufs
+{
+	return 0;
+	(void)nl;
+	(void)type;
+	(void)buf;
+	(void)len;
+}
+
+unsigned int mnl_socket_get_portid(const struct mnl_socket *nl)
+{
+	return 0;
+	(void)nl;
+}
+
+// configuration packet ready
+ssize_t mnl_socket_sendto(const struct mnl_socket *hconst,
+					const void *buf, size_t len)
+{
+	struct my_params *par = (struct my_params*)buf;
+	int num = par->queue_num;
 	if (num >= TEST_QUEUE_MAX)
 	{
 		fprintf(stderr,
 			MY_PREFIX ": cannot create queue %d: only support up to %d\n",
 			num, TEST_QUEUE_MAX);
-		return NULL;
-	}
-	
-	if (h->q[num].h == h) // already opened. Is this an error?
-	{
-		return NULL;
-	}
-	
-	h->q[num].h = h;
-	h->q[num].cb = cb;
-	h->q[num].data = data;
-	
-	return &h->q[num];
-}
-
-static int get_queue_num(struct nfq_q_handle *qh)
-{
-	struct nfq_handle *h = qh->h;
-	unsigned num = qh - (struct nfq_q_handle *)h;
-	
-	if (num >= TEST_QUEUE_MAX || h != h->q[num].h)
-	{
-		fputs(
-			MY_PREFIX ": bad handle!!\n",
-			stderr);
-		assert(0);
 		return -1;
 	}
 
-	return num;
+	struct mnl_socket *h = (struct mnl_socket*) hconst;
+	if (par->cmd == my_NFQNL_CFG_CMD_BIND)
+	{
+		if (h->q[num] == num) // already opened. Is this an error?
+		{
+			return -1;
+		}
+
+		h->q[num] = num;
+		return len;
+	}
+
+	if (par->cmd == my_NFQNL_CFG_CMD_UNBIND)
+	{
+	}
+
+	if (par->cmd == my_NFQNL_COPY_PACKET) // range
+	{
+		
+	}
+
+	if (par->cmd == my_NFQNL_MSG_VERDICT)
+	{
+		if (par->mark)
+			print_out(h,
+				"nfq_set_verdict_mark: %d, %d, %d\n",
+				par->id, par->verdict, par->mark);
+		else
+			print_out(h,
+				"nfq_set_verdict: %d, %d\n",
+				par->id, par->verdict);
+		return len;
+	}
+
+	if (par->cmd == my_NFQNL_MSG_CONFIG_MAXLEN)
+	{
+		print_out(h, "nfq_set_queue_maxlen: %u\n", par->queue_maxlen);
+	}
+	return len;
 }
 
-/* unbind this socket from a specific queue number */
-int nfq_destroy_queue(struct nfq_q_handle *qh)
+
+
+int mnl_cb_run(const void *buf_void, size_t numbytes, unsigned int seq,
+			     unsigned int portid, mnl_cb_t cb_data, void *data)
 {
-	if (get_queue_num(qh) < 0)
-		return -1;
 
-	memset(qh, 0, sizeof *qh);
-	return 0;
+	unsigned char *buf = (unsigned char*)buf_void;
+	int rtc = -1;
+	while (numbytes >= sizeof(struct nfq_data))
+	{
+		struct nfq_data *pkt = (struct nfq_data*)buf;
+		for (int q = 0; q < TEST_QUEUE_MAX; ++q)
+			if (pkt->queue[q])
+			{
+				if (pkt->nl->q[q] != q) // not opened?
+				{
+					fprintf(stderr,
+						MY_PREFIX ": queue %d not opened\n", q);
+					return -1;
+				}
+
+				pkt->current_q = q;
+				rtc = (*cb_data)((const struct nlmsghdr *)buf, data);
+				if (rtc <= 0)
+				{
+					suicide();
+					return rtc;
+				}
+			}
+
+		buf += sizeof(struct nfq_data);
+		numbytes -= sizeof(struct nfq_data);
+	}
+
+	return rtc;
+	(void)seq;
+	(void)portid;
 }
 
-int nfq_handle_packet(struct nfq_handle *h, char *buf, int len)
+#define MAX_NEXTBUFS 10
+static void *to_be_freed[MAX_NEXTBUFS];
+static unsigned int cnt_to_be_freed;
+static void add_to_be_freed(void*n)
+{
+	if (cnt_to_be_freed < MAX_NEXTBUFS)
+		to_be_freed[cnt_to_be_freed++] = n;
+}
+
+static void free_nextbuf(void)
+{
+	while (cnt_to_be_freed > 0)
+		free(to_be_freed[--cnt_to_be_freed]);
+}
+
+ssize_t mnl_socket_recvfrom(const struct mnl_socket *nlconst,
+					  void *pktbuf_void, size_t pktlen)
 {	
 	test_cmd cmd;
 	int rtc = 0, i;
-	
-	struct nfq_data *pkt = &h->next_pkt;
-	
-	h->sv_sent = 0;
-	
-	while (len > 0 && (unsigned)len >= sizeof cmd && rtc >= 0)
+
+	struct mnl_socket *nl = (struct mnl_socket*) nlconst;
+	struct nfq_data *pkt = &nl->next_pkt;
+
+	if (nl->sv_sent == 0)
+	{
+		suicide();
+		return 0;
+	}
+
+	unsigned char *pktbuf = pktbuf_void;
+	ssize_t pktlen_copied = 0;
+
+	unsigned char *buf = nl->nextbuf;
+	add_to_be_freed(buf);
+	size_t len = nl->nextbuf_sz;
+	nl->sv_sent = 0;
+	nl->nextbuf = NULL;
+	nl->nextbuf_sz = 0;
+
+	while (len > 0 && len >= sizeof cmd && rtc >= 0)
 	{
 		int pkts = 0;
-		
+
 		memcpy(&cmd, buf, sizeof cmd);
 		buf += sizeof cmd;
 		len -= sizeof cmd;
-			
+
 		switch (cmd.cmd)
 		{
 			case test_cmd_group:
@@ -836,67 +1014,65 @@ int nfq_handle_packet(struct nfq_handle *h, char *buf, int len)
 				break;
 
 			case test_cmd_sleep:
-				print_out(h, "sleep %d\n", cmd.arg);
+				print_out(nl, "sleep %d\n", cmd.arg);
 				sleep(cmd.arg);
 				break;
 
 			case test_cmd_settime:
-				time(&h->current);
-				// print_out(h, "------------->reset time to ...... %ld\n",
-				//	(long)h->current);
+				time(&nl->current);
+				// print_out(nl, "------------->reset time to ...... %ld\n",
+				//	(long)nl->current);
 				break;
 
 			case test_cmd_putenv:
 			{
 				assert(cmd.arg >= 0 && cmd.arg < TEST_ARGS_MAX);
-				assert(h->cmd_arg[cmd.arg] != NULL);
-				
-				char *arg = simple_expand(h->cmd_arg[cmd.arg]);
+				assert(nl->cmd_arg[cmd.arg] != NULL);
+
+				char *arg = simple_expand(nl->cmd_arg[cmd.arg]);
 				if (arg == NULL)
 				{
-					close_socketpair(h);
 					rtc = -1;
 				}
 
-				// print_out(h, "putenv %s\n", arg);
+				// print_out(nl, "putenv %s\n", arg);
 				if (putenv(arg) != 0)
 					fprintf(stderr,
 						"%s: putenv %s failed: %s\n",
 						my_prefix, arg, strerror(errno));
 
 				// don't free arg
-				h->cmd_arg[cmd.arg] = NULL;
+				nl->cmd_arg[cmd.arg] = NULL;
 				break;
 			}
 
 			case test_cmd_system:
 			{
 				assert(cmd.arg >= 0 && cmd.arg < TEST_ARGS_MAX);
-				assert(h->cmd_arg[cmd.arg] != NULL);
-				print_out(h, "system %s\n", h->cmd_arg[cmd.arg]);
+				assert(nl->cmd_arg[cmd.arg] != NULL);
+				print_out(nl, "system %s\n", nl->cmd_arg[cmd.arg]);
 
-				int rc = system(h->cmd_arg[cmd.arg]);
+				int rc = system(nl->cmd_arg[cmd.arg]);
 				if (rc == -1)
 					fprintf(stderr,
 						"%s: system %s failed: %s\n",
 						my_prefix,
-						h->cmd_arg[cmd.arg],
+						nl->cmd_arg[cmd.arg],
 						strerror(errno));
 				else if (WIFSIGNALED(rc) &&
 					(WTERMSIG(rc) == SIGINT || WTERMSIG(rc) == SIGQUIT))
 				{
-					close_socketpair(h);
 					rtc = -1;
 				}
 				else if (WEXITSTATUS(rc))
 					fprintf(stderr,
 						"%s: system %s exited with status %d\n",
 						my_prefix,
-						h->cmd_arg[cmd.arg],
+						nl->cmd_arg[cmd.arg],
 						WEXITSTATUS(rc));
 
-				free(h->cmd_arg[cmd.arg]);
-				h->cmd_arg[cmd.arg] = NULL;
+				free(nl->cmd_arg[cmd.arg]);
+				nl->cmd_arg[cmd.arg] = NULL;
 				break;
 			}
 
@@ -906,160 +1082,50 @@ int nfq_handle_packet(struct nfq_handle *h, char *buf, int len)
 		}
 
 		for (i = 0; i < pkts && rtc >= 0; ++i)
-		{			
-			if (!h->have_next_pkt)
-				if (read_test_input(h))
+		{
+			if (!nl->have_next_pkt)
+				if (read_test_input(nl))
 					break;
 
-			assert(h->have_next_pkt);
-			h->have_next_pkt = 0;
+			assert(nl->have_next_pkt);
+			nl->have_next_pkt = 0;
 			
 			// increase current --ibd-judge gets the delta specified on pkt
-			h->current += h->next_pkt.timestamp;
-			h->next_pkt.timestamp = h->current;
-			
-			// call all registered callbacks		
-			for (int q = 0; q < TEST_QUEUE_MAX && rtc >= 0; ++q)
-				if (h->next_pkt.queue[q] && h->q[q].cb)
-					rtc = h->q[q].cb(&h->q[q], NULL /* ?? */, pkt, h->q[q].data);
+			nl->current += nl->next_pkt.timestamp;
+			nl->next_pkt.timestamp = nl->current;
 
-			memset(pkt, 0, sizeof *pkt);
-		}		
+			// copy the packet to ibd-judge's buffer
+			pkt->nl = nl;
+			if (pktlen > sizeof *pkt)
+			{
+				memcpy(pktbuf, pkt, sizeof *pkt);
+				pktbuf += sizeof *pkt;
+				pktlen -= sizeof *pkt;
+				pktlen_copied += sizeof *pkt;
+
+				memset(pkt, 0, sizeof *pkt);
+			}
+			else
+			{
+				fprintf(stderr,
+					MY_PREFIX ": buffer too small: %zu, needed %zu\n",
+					len, nl->nextbuf_sz);
+				return -1;
+			}
+		}
 	}
 
-	/*
-	* make sure that either socketpair is closed
-	* or there is something to be read from it
-	*/	
-	if (h->sv_open && !h->sv_sent)
+	free_nextbuf();
+
+	if (nl->in == NULL)
+		suicide();
+	else if (!nl->sv_sent)
 	{
-		if (h->have_next_pkt)
-			send_noop_packet(h);
-		else
-			read_test_input(h);
+		if (!nl->have_next_pkt)
+			read_test_input(nl);
 	}
-		
-	return rtc < 0? rtc: 0;
+
+	return rtc < 0? rtc: pktlen_copied;
 }
 
-int nfq_set_mode(struct nfq_q_handle *qh, u_int8_t mode, u_int32_t range)
-{
-	if (get_queue_num(qh) < 0)
-		return -1;
-
-	print_out(qh->h, "nfq_set_mode: %d, %d\n", mode, range);
-	return 0;
-}
-
-int nfq_set_queue_maxlen(struct nfq_q_handle *qh, u_int32_t queue_maxlen)
-{
-	if (get_queue_num(qh) < 0)
-		return -1;
-
-	print_out(qh->h, "nfq_set_queue_maxlen: %u\n", queue_maxlen);
-	return 0;
-}
-
-int nfq_set_verdict2(struct nfq_q_handle *qh,
-                            u_int32_t id,
-                            u_int32_t verdict, 
-                            u_int32_t mark,
-                            u_int32_t datalen,
-                            const unsigned char *buf)
-{
-	if (get_queue_num(qh) < 0)
-		return -1;
-	print_out(qh->h,
-		"nfq_set_verdict_mark: %d, %d, %d\n",
-		id, verdict, mark);
-	return 0;
-	(void)datalen;
-	(void)buf;
-}
-
-int nfq_set_verdict(struct nfq_q_handle *qh,
-                             u_int32_t id,
-                             u_int32_t verdict,
-                             u_int32_t data_len,
-                             const unsigned char *buf)
-{
-	if (get_queue_num(qh) < 0)
-		return -1;
-	print_out(qh->h,
-		"nfq_set_verdict: %d, %d\n",
-		id, verdict);
-	return 0;
-	(void)data_len;
-	(void)buf;
-}
-
-/*************************************************************
- * Message parsing functions 
- *************************************************************/
-
-struct nfqnl_msg_packet_hdr *nfq_get_msg_packet_hdr(struct nfq_data *pkt)
-{
-	return &pkt->hdr;
-}
-
-uint32_t nfq_get_nfmark(struct nfq_data *pkt)
-{
-	(void)pkt;
-	return 0;
-}
-
-int nfq_get_timestamp(struct nfq_data *pkt, struct timeval *tv)
-{
-	if (pkt->timestamp == (time_t)-1)
-		return -1;
-
-	tv->tv_sec = pkt->timestamp;
-	tv->tv_usec = 0;
-
-	return 0;
-}
-
-/* all nfq_get_*dev() functions return 0 if not set, since linux only allows
- * ifindex >= 1, see net/core/dev.c:2600  (in 2.6.13.1) */
-u_int32_t nfq_get_indev(struct nfq_data *pkt)
-{
-	(void)pkt;
-	return 0;
-}
-
-u_int32_t nfq_get_physindev(struct nfq_data *pkt)
-{
-	(void)pkt;
-	return 0;
-}
-
-u_int32_t nfq_get_outdev(struct nfq_data *pkt)
-{
-	(void)pkt;
-	return 0;
-}
-
-u_int32_t nfq_get_physoutdev(struct nfq_data *pkt)
-{
-	(void)pkt;
-	return 0;
-}
-
-struct nfqnl_msg_packet_hw *nfq_get_packet_hw(struct nfq_data *pkt)
-{
-	(void)pkt;
-	return 0;
-}
-
-int nfq_get_payload(struct nfq_data *pkt, unsigned char **data)
-{
-	*data = pkt->packet.payload;
-	return sizeof pkt->packet.payload;
-}
-
-unsigned int nfnl_rcvbufsiz(const struct nfnl_handle *h, unsigned int size)
-{
-	(void)h;
-	return size*2;
-}
 
